"""Code for metadata regarding finished calculation results."""

# General imports.
from datetime import timedelta, datetime
import math
import itertools
from deepmerge import conservative_merger
from pathlib import Path
import copy
import warnings
from scipy import integrate

from digichem.misc.time import latest_datetime, total_timedelta, date_to_string,\
    timedelta_to_string
from digichem.misc.text import andjoin
from digichem.exception import Result_unavailable_error
from digichem.result import Result_object
import digichem
from digichem import translate
from digichem.memory import Memory


class Solvent(Result_object):
    """
    Class for storing solvent metadata.
    """
    
    def __init__(self, model = None, name = None, params = None):
        self.model = model
        self.name = name
        # A raw dictionary of solvent related parameters.
        self.params = params if params is not None else {}
        
        # If we are missing only one of name and params['epsilon'], look up the missing value.
        if self.name is None and "epsilon" in self.params:
            try:
                self.name = translate.Solvent.epsilon_to_name(self.params['epsilon'])
            
            except ValueError:
                # Nothing close.
                pass
        
        if self.name is not None:    
            self.name = self.name.capitalize()
        
    @classmethod
    def from_parser(self, parser):
        """
        Construct a Solvent object from an output file parser.
        
        :param parser: Output data parser.
        :return: A populated Solvent object.
        """
        return self(
            name = parser.data.metadata.get('solvent_name', None),
            model = parser.data.metadata.get('solvent_model', None),
            params = parser.data.metadata.get('solvent_params', {}),
        )
        
    @property
    def description(self):
        """
        A common-name description of this solvent.
        """
        # If we have no solvent, say so.
        if self.model is None:
            return "Gas-phase"
        
        # If we have a name, just use that.
        elif self.name is not None:
            return self.name
        
        # If we only have epsilon, use that.
        elif "epsilon" in self.params:
            return "ε = {}".format(self.params['epsilon'])
        
        # Give up.
        else:
            return "Unknown"
        
    def _dump_(self, digichem_options, all):
        return {
            "model": self.model,
            "name": self.name,
            "params": self.params
        }
        
    @classmethod
    def merge(self, *multiple_objects):
        """
        Merge multiple solvent implementations together.
        """
        # Check all items are the same.
        if not all(obj == multiple_objects[0] for obj in multiple_objects if obj is not None):
            warnings.warn("Refusing to merge different solvent methods")
            return self()
            
        return multiple_objects[0]
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(
            model = data.get('model', None),
            name = data.get('name', None),
            params = data.get('params', {})
        )
        
    def __eq__(self, other):
        """Is this solvent implementation the same as another one?"""
        return (
            # Might want to do this case-insensitive.
            self.model == other.model and
            abs(self.params.get("epsilon", math.inf) - self.params.get("epsilon", math.inf)) < 0.001
        )


class Metadata(Result_object):
    """
    Class for storing calculation metadata.
    """
    
    # A dictionary of known computational orders, where the key is name of the method (upper case) and the value is an integer describing the ordered 'level of theory'.
    METHODS = {
        "HF":       1,
        "DFT":      2,
        "LMP2":     3,
        "DF-MP2":   4,
        "MP2":      5,
        "MP3":      6,
        "MP4":      7,
        "MP5":      8,
        "CCSD":     9,
        "CCSD(T)":  10,
        "CCSD-T":   11
    }
    
    def __init__(
            self,
            jobId = None,
            name = None,
            user = None,
            log_files = None,
            auxiliary_files = None,
            history = None,
            date = None,
            insert_date = None,
            duration = None,
            package = None,
            package_version = None,
            calculations = None,
            success = None,
            methods = None,
            functional = None,
            basis_set = None,
            charge = None,
            multiplicity = None,
            optimisation_converged = None,
            temperature = None,
            pressure = None,
            orbital_spin_type = None,
            digichem_version = None,
            silico_version = None,
            solvent = None,
            num_cpu = None,
            memory_available = None,
            memory_used = None,
            performance = None,
            
            # Deprecated.
            solvent_model = None,
            solvent_name = None,):
        """
        Constructor for result Metadata objects.
        
        :param jobId: If this result was generated from a digichem calculation, the relevant jobID.
        :param name: Optional name of this calculation result.
        :param user: The username of the user who parsed this result.
        :param log_files: An optional list of text-based calculation log files from which this result was parsed.
        :param auxiliary_files: An optional dict of auxiliary files associated with this calculation result.
        :param history: Optional SHA of the calculation from which the coordinates of this calculation were generated.
        :param num_calculations: Optional number of individual calculations this metadata represents.
        :param date: Optional date (datetime object) of this calculation result.
        :param insert_date: Optional date (datetime object) of when this calculation result was stored (normally in a DB).
        :param duration: Optional duration (timedelta object) of this calculation.
        :param package: Optional string identifying the computational chem program that performed the calculation.
        :param package_version: Optional string identifying the version of the computational chem program that performed the calculation.
        :param calculations: A list of strings of the different calculations carried out (Opt, Freq, TD, TDA, SP etc).
        :param success: Whether the calculation completed successfully or not.
        :param methods: List of methods (DFT, HF, MPn etc) used in the calculation.
        :param functional: Functional used in the calculation.
        :param basis_set: Basis set used in the calculation.
        :param charge: Charge (positive or negative integer) of the system studied.
        :param multiplicity: The multiplicity of the system.
        :param optimisation_converged: Whether the optimisation converged or not.
        :param temperature: The temperature used in the calculation (not always relevant).
        :param pressure: The pressure used in the calculation (not always relevant).
        :param orbital_spin_type: The types of orbitals that have been calculated, either 'restricted' or 'unrestricted'.
        :param num_cpu: The number of CPUs used to perform this calculation.
        :param memory_available: The maximum amount of memory available to this calculation (the amount requested by the user).
        :param memory_used: The maximum amount of memory used by the calculation (the amount requested by the user).
        """
        self.jobId = jobId
        self.num_calculations = 1
        self.name = name
        self.user = user
        self.log_files = log_files if log_files is not None else []
        self.auxiliary_files = auxiliary_files if auxiliary_files is not None and len(auxiliary_files) != 0 else {}
        self.history = history
        self.date = date
        self.insert_date = insert_date
        self.duration = duration
        self.package = package
        self.package_version = package_version
        self.calculations = calculations if calculations is not None else []
        self.success = success
        self.methods = methods if methods is not None else []
        self.functional = functional
        self.basis_set = basis_set
        # TODO: charge and mult should be deprecated here, they are available in ground_state.
        self.charge = charge
        self.multiplicity = multiplicity
        self.optimisation_converged = optimisation_converged
        self.temperature = temperature
        self.pressure = pressure
        self.orbital_spin_type = orbital_spin_type
        # TOOD: Ideally this would be parsed from the calculation output somehow, but this is fine for now.
        if silico_version:
            self.digichem_version = silico_version
        self.digichem_version = digichem.__version__ if digichem_version is None else digichem_version
        self.solvent = solvent
        self.num_cpu = num_cpu
        self.memory_available = memory_available
        self.memory_used = memory_used
        self.performance = performance
        
        # Deprecated solvent system.
        if solvent_model is not None:
            self.solvent.model = solvent_model
        
        if solvent_name is not None:
            self.solvent.name = solvent_name
    
    # TODO: This is more than a bit clumsy and in general the handling of names should be improved.
    @property
    def molecule_name(self):
        return Path(self.name).name if self.name is not None else None
    
    @property
    def level_of_theory(self):
        """
        A short-hand summary of the methods and basis sets used.
        """
        theories = []
        if len(self.converted_methods) > 0:
            #theories.extend(self.converted_methods)
            theories.append(self.converted_methods[-1])
            
        if self.basis_set is not None:
            theories.append(self.basis_set)
            
        return("/".join(theories))
        
    @classmethod
    def merge(self, *multiple_metadatas):
        """
        Merge multiple metadata objects into a single metadata.
        """
        return Merged_metadata.from_metadatas(*multiple_metadatas)
    
    @property
    def converted_methods(self):
        """
        Similar to the methods attribute but where DFT is replaced with the actual functional used.
        """
        return [self.functional if method == "DFT" and self.functional is not None else method for method in self.methods if method is not None]
    
    @property
    def description(self):
        desc = []
        if self.name is not None:
            desc.append(self.name)
        
        desc.append(self.identity_string)
        
        return ", ".join(desc)
        
    @property
    def identity(self):
        """
        A dictionary of critical attributes that identify a calculation.
        """
        # Get our list of methods, replacing 'DFT' with the functional used if available.
        return {
            "package": self.package,
            "calculations": self.calculations_string,
            "methods": ", ".join(self.converted_methods) if len(self.converted_methods) > 0 else None,
            "basis": self.basis_set,
        }
        
    @property
    def identity_string(self):
        """
        A string that identifies this calculation.
        """
        return " ".join([value for value in self.identity.values() if value is not None])
        
    
    @classmethod
    def sorted_methods(self, methods):
        """
        Order a list of methods (HF, DFT, MP2 etc) in terms of 'level of theory', with the lowest level (HF or DFT probably) first and highest (MP4, CCSD etc) last.
        """
        return sorted((method.upper() for method in methods), key = lambda method: self.METHODS[method] if method in self.METHODS else math.inf)
    
    @property
    def package_string(self):
        """
        The comp chem package and version combined into one string.
        """
        package_string = self.package if self.package is not None else ""
        
        # Add version string if we have it.
        package_string += " " if package_string != "" else ""
        package_string += "({})".format(self.package_version)
        
        # Done.
        return package_string 
        
    @property
    def calculations_string(self):
        """
        Get the list of calculation types as a single string, or None if there are no calculations set.
        """
        return ", ".join(self.calculations) if len(self.calculations) != 0 else None
    
    @property
    def methods_string(self):
        """
        Get the list of calculation methods as a single string, or None if there are no methods set.
        """
        return ", ".join(self.methods) if len(self.methods) != 0 else None
    
    @property
    def human_calculations_string(self):
        """
        A version of calculations_string that is more pleasant to read.
        """
        calculations = []
        if "Single Point" in self.calculations:
            calculations.append("single point energy")
        
        if "Optimisation" in self.calculations:
            calculations.append("optimised structure")
            
        if "Frequencies" in self.calculations:
            calculations.append("vibrational frequencies")
            
        if "Excited States" in self.calculations:
            calculations.append("excited states")
            
        if "NMR" in self.calculations:
            calculations.append("NMR properties")
            
        if len(calculations) == 0:
            # Use a generic term.
            calculations.append("properties")
        
        return andjoin(calculations)
            
    @classmethod
    def get_calculation_types_from_cclib(self, ccdata):
        """
        Determine what types of calculation we did from the data provided by cclib.
        
        :param ccdata: The data from cclib.
        :return: A list of strings.
        """
        calcs = []
        # Need to think of a better way to determine what is an optimisation, as frequency calcs contain an optdone for some reason.
        #if hasattr(ccdata, 'optdone'):
        if len(getattr(ccdata, 'scfenergies', [])) > 1:
            calcs.append('Optimisation')
        if hasattr(ccdata, 'vibfreqs'):
            calcs.append('Frequencies')
        if hasattr(ccdata, 'etenergies'):
            calcs.append('Excited States')
        if hasattr(ccdata, 'nmrtensors'):
            calcs.append('NMR')
        # If our list is empty, assume we did an SP.
        if len(calcs) == 0 and ( hasattr(ccdata, 'scfenergies') or hasattr(ccdata, 'mpenergies') or hasattr(ccdata, 'ccenergies') ):
                calcs = ['Single Point']
        # Return our list.
        return calcs
    
    @classmethod
    def get_methods_from_cclib(self, ccdata):
        """
        Get a list of method types (DFT, MP etc.) from the data provided by cclib.
        
        :param ccdata: The data from cclib.
        :return: A list of strings of the methods used (each method appears once only; the order has no special significance). The list may be empty.
        """
        return self.sorted_methods((set(ccdata.metadata.get('methods', []))))
        #return list(dict.fromkeys(ccdata.metadata.get('methods', [])))
    
    @classmethod
    def get_orbital_spin_type_from_cclib(self, ccdata):
        """
        Determine the orbital type (restricted, unrestricted) from the data provided by cclib.
        
        :return ccdata: The data from cclib.
        :return: A string describing the orbital type ('restricted', 'unrestricted' or 'none').
        """
        #mo_len = len(ccdata.get('moenergies', []))
        mo_len = len(getattr(ccdata, 'moenergies', []))
        if mo_len == 1:
            return "restricted"
        elif mo_len == 2:
            return "unrestricted"
        else:
            return "none"
    
    @classmethod
    def from_parser(self, parser):
        """
        Construct a Metadata object from an output file parser.
        
        :param parser: Output data parser.
        :return: A populated Metadata object.
        """
        try:
            # Do some processing on time objects.
            if parser.data.metadata.get('wall_time', None) is not None:
                duration = sum(parser.data.metadata['wall_time'], timedelta())
                
            elif parser.data.metadata.get('cpu_time', None) is not None and parser.data.metadata.get('num_cpus', None) is not None:
                # We can estimate the wall time from the CPU time and number of cpus.
                duration = sum(parser.data.metadata['cpu_time'], timedelta()) / parser.data.metadata['num_cpus']
                
            else:
                duration = None
                
            if parser.data.metadata.get('date', None) is not None:
                date = datetime.fromtimestamp(parser.data.metadata['date'])
            else:
                date = None
                
            memory_used = Memory(parser.data.metadata['memory_used']) if "memory_used" in parser.data.metadata else None
            memory_available = Memory(parser.data.metadata['memory_available']) if "memory_available" in parser.data.metadata else None
            
            # TODO: This doesn't seem to make sense; the parser already contains a metadata object...
            return self(
                jobId = parser.data.metadata.get('jobId', None),
                name = parser.data.metadata.get('name', None),
                user = parser.data.metadata.get('user', None),
                log_files = parser.data.metadata.get('log_files', None),
                auxiliary_files = parser.data.metadata.get('auxiliary_files', None),
                date = date,
                duration = duration,
                package = parser.data.metadata.get('package', None),
                package_version = parser.data.metadata.get('package_version', None),
                calculations = self.get_calculation_types_from_cclib(parser.data),
                success = parser.data.metadata.get('success', None),
                methods = self.get_methods_from_cclib(parser.data),
                functional = parser.data.metadata.get('functional', None),
                basis_set = parser.data.metadata.get('basis_set', None),
                
                charge = getattr(parser.data, 'charge', None),
                multiplicity = getattr(parser.data, 'mult', None),
                optimisation_converged = getattr(parser.data, 'optdone', None),
                temperature = getattr(parser.data, 'temperature', None),
                pressure = getattr(parser.data, 'pressure', None),
                orbital_spin_type = self.get_orbital_spin_type_from_cclib(parser.data),
                
                solvent = Solvent.from_parser(parser),
                
                num_cpu = parser.data.metadata.get('num_cpu', None),
                memory_available = memory_available,
                memory_used = memory_used,
                performance = Performance.from_parser(parser) if "performance" in parser.data.metadata else None
            )
        except AttributeError:
            # There is no metadata available, give up.
            raise Result_unavailable_error("Metadata", "no metadata is available")
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        # Most attributes we can just dump as is.
        attr_dict = {
            "name": self.name,
            "log_files": [str(log_file) for log_file in self.log_files],
            "auxiliary_files": {aux_file_name: str(aux_file) for aux_file_name, aux_file in self.auxiliary_files.items()}
        }
        
        attrs = [
            "jobId",
            "history",
            "charge",
            "multiplicity",
            "user",
            "package",
            "package_version",
            "digichem_version",
            "calculations",
            "methods",
            "functional",
            "basis_set",
            "orbital_spin_type",
            "success",
            "optimisation_converged",
        ]
        attr_dict.update({attr: getattr(self, attr) for attr in attrs})
        
        # Add some more complex stuff.
        attr_dict['date'] = {
            "value": self.date.timestamp() if self.date is not None else None,
            "units": "s",
            "string": date_to_string(self.date) if self.date is not None else None
        }
        attr_dict['insert_date'] = {
            "value": self.insert_date.timestamp() if self.insert_date is not None else None,
            "units": "s",
            "string": date_to_string(self.insert_date) if self.insert_date is not None else None
        }
        attr_dict['duration'] = {
            "value": self.duration.total_seconds() if self.duration is not None else None,
            "units": "s",
            "string": timedelta_to_string(self.duration) if self.duration is not None else None
        }
        attr_dict["temperature"] = {
            "value": self.temperature,
            "units": "K"
        }
        attr_dict["pressure"] = {
            "value": self.pressure,
            "units": "atm"
        }
        attr_dict["solvent"] = self.solvent.dump(digichem_options, all)
        
        attr_dict['num_cpu'] = self.num_cpu
        for attr_name in ("memory_used", "memory_available"):
            if getattr(self, attr_name) is not None:
                value, unit = getattr(self, attr_name).auto_units
                attr_dict[attr_name] = {
                    "value": value,
                    "units": unit
                }
            else:
                attr_dict[attr_name] = {
                    "value": None,
                    "units": None
                }
        
        attr_dict['performance'] = self.performance.dump(digichem_options, all) if self.performance else None

        return attr_dict
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        # Assemble our args to pass to the constructor.
        # Most of these can be used from the metadata dict as is.
        kwargs = copy.deepcopy(data)
        
        # For more complex fields, use the data item.
        for attr in ['insert_date', 'date', 'duration', 'temperature', "pressure"]:
            if attr in data:
                kwargs[attr] = data[attr]['value']
            
            else:
                kwargs[attr] = None
        
        kwargs['insert_date'] = datetime.fromtimestamp(kwargs['insert_date']) if kwargs['insert_date'] is not None else None
        kwargs['date'] = datetime.fromtimestamp(kwargs['date']) if kwargs['date'] is not None else None
        kwargs['duration'] = timedelta(seconds = kwargs['duration'])  if kwargs['duration'] is not None else None
        
        kwargs['solvent'] = Solvent.from_dump(data.get('solvent', {}), result_set, options)
        
        for attr_name in ("memory_used", "memory_available"):
            if attr_name in data and data[attr_name]['value'] is not None:
                kwargs[attr_name] = Memory.from_units(data[attr_name]["value"], data[attr_name]["units"])
            
            else:
                kwargs[attr_name] = None

        kwargs['performance'] = Performance.from_dump(data['performance'], result_set, options) if "performance" in data and data['performance'] is not None else None
        
        return self(**kwargs)


class Merged_metadata(Metadata):
    """
    A modified metadata class for merged calculation results.
    """
    
    
    def __init__(self, num_calculations, *args, **kwargs):
        """
        :param num_calculations: The number of merged calculations this metadata represents.
        """
        super().__init__(*args, log_files = None, auxiliary_files = None, **kwargs)
        self.num_calculations = num_calculations
        
    @classmethod
    def from_metadatas(self, *multiple_metadatas):
        """
        Create a merged metadata object from multiple metadata objects.
        
        :param multiple_metadatas: A list of metadata objects to merge.
        :returns: A new Merged_metadata object.
        """
        # Our merged metadata.
        merged_metadata = self(num_calculations = len(multiple_metadatas))
        for attr in ("name", "user", "package", "package_version", "functional", "basis_set"):
            setattr(merged_metadata, attr, self.merged_attr(attr, multiple_metadatas))
            
        # We take the latest of the two dates.
        merged_metadata.date = latest_datetime(*[metadata.date for metadata in multiple_metadatas if metadata.date is not None])
        # And the total duration.
        merged_metadata.duration = total_timedelta(*[metadata.duration for metadata in multiple_metadatas if metadata.duration is not None])
        
        # Merge methods and calculations (but keep unique only).
        merged_metadata.calculations = list(dict.fromkeys(itertools.chain(*(metadata.calculations for metadata in multiple_metadatas))))
        merged_metadata.methods = self.sorted_methods(set(itertools.chain(*(metadata.methods for metadata in multiple_metadatas))))
        
        # Keep the solvent if it's the same for all, otherwise discard.
        merged_metadata.solvent = multiple_metadatas[0].solvent.merge(*[other.solvent for other in multiple_metadatas[1:]])
        
        # Combine performance data.
        merged_metadata.num_cpu = sum([meta.num_cpu for meta in multiple_metadatas if meta.num_cpu is not None])
        merged_metadata.memory_available = Memory(sum([int(meta.memory_available) for meta in multiple_metadatas if meta.memory_available is not None]))
        merged_metadata.memory_used = Memory(sum([int(meta.memory_used) for meta in multiple_metadatas if meta.memory_used is not None]))
        
        # We are only successful if all calcs are successful.
        merged_metadata.success = all((metadata.success for metadata in multiple_metadatas))
        converged = [metadata.optimisation_converged for metadata in multiple_metadatas]
        merged_metadata.optimisation_converged = False if False in converged else True if True in converged else None
        
        # Only keep these if all the same.
        for attr in ("temperature", "pressure"):
            # Get a unique list of the attrs.
            attr_set = set(getattr(metadata, attr) for metadata in multiple_metadatas)
            
            # Only keep if we have exactly one entry.
            if len(attr_set) == 1:
                setattr(merged_metadata, attr, tuple(attr_set)[0])
                
        # We take the first orbital spin type charge and mult, as this is what our orbital list will actually be.
        merged_metadata.orbital_spin_type = multiple_metadatas[0].orbital_spin_type
        merged_metadata.charge = multiple_metadatas[0].charge
        merged_metadata.multiplicity = multiple_metadatas[0].multiplicity
        
        # CAREFULLY merge aux files, so that later files do not overwrite earlier ones.
        # This is useful behaviour because it matches how other results are merged, so certain aux files
        # (turbomole density files) will still match their respective results.
        for metadata in multiple_metadatas:
            merged_metadata.auxiliary_files = conservative_merger.merge(merged_metadata.auxiliary_files, metadata.auxiliary_files)
        
        return merged_metadata

class Performance(Result_object):
    """
    Performance metrics and profiling data for the calculation.
    """

    def __init__(
        self,
        duration = [],
        memory_used = [],
        memory_used_percent = [],
        memory_available = [],
        memory_available_percent = [],
        cpu_used = [],
        output_available = [],
        scratch_used = [],
        scratch_available = [],
        memory_allocated = None,
        cpu_allocated = None,
    ):
        self.duration = duration
        self.memory_used = memory_used
        self.memory_used_percent = memory_used_percent
        self.memory_available = memory_available
        self.memory_available_percent = memory_available_percent
        self.cpu_used = cpu_used
        self.output_available = output_available
        self.scratch_used = scratch_used
        self.scratch_available = scratch_available

        self.memory_allocated = memory_allocated if memory_allocated is not None else self.max_mem
        self.cpu_allocated = cpu_allocated if cpu_allocated is not None else math.ceil(max(cpu_used) / 100)

    @property
    def output_space(self):
        warnings.warn("output_space is deprecated, use output_available instead", DeprecationWarning)
        return self.output_available

    @property
    def scratch_space(self):
        warnings.warn("scratch_space is deprecated, use scratch_available instead", DeprecationWarning)
        return self.scratch_available
    

    @classmethod
    def from_parser(self, parser):
        """
        Construct a Performance object from an output file parser.
        
        :param parser: Output data parser.
        :return: A populated Performance object.
        """
        return self(
            duration = parser.data.metadata['performance']['duration'].tolist(),
            memory_used = parser.data.metadata['performance']['memory_used'].tolist(),
            memory_allocated = Memory(parser.data.metadata['memory_available']) if "memory_available" in parser.data.metadata else None,
            memory_used_percent = parser.data.metadata['performance']['memory_used_percent'].tolist(),
            memory_available = parser.data.metadata['performance']['memory_available'].tolist(),
            memory_available_percent = parser.data.metadata['performance']['memory_available_percent'].tolist(),
            cpu_used = parser.data.metadata['performance']['cpu_used'].tolist(),
            cpu_allocated = parser.data.metadata.get('num_cpu', None),
            output_available = parser.data.metadata['performance']['output_available'].tolist(),
            scratch_used = parser.data.metadata['performance']['scratch_used'].tolist() if 'scratch_used' in parser.data.metadata['performance'] else [0] * len(parser.data.metadata['performance']['duration']),
            scratch_available = parser.data.metadata['performance']['scratch_available'].tolist()
        )
    
    @property
    def max_mem(self):
        """
        The maximum amount of memory used in the calculation (in bytes)
        """
        return max(self.memory_used)
    
    @property
    def memory_margin(self):
        max_memory = float(self.memory_allocated) if self.memory_allocated is not None else self.max_mem
        
        return max_memory - self.max_mem
    
    @property
    def memory_efficiency(self):
        """
        Calculate the memory efficiency of this calculation.

        :param max_memory: The amount of allocated memory (in bytes), this will be guestimated automatically if not available.
        """
        # Integrate to find the number of byte seconds used.
        area = integrate.trapezoid(self.memory_used, self.duration)

        # How much we should/could have used.
        total_area = (self.duration[-1] - self.duration[0]) * float(self.memory_allocated)

        # Return as %.
        try:
            return area / total_area * 100
        
        except Exception:
            return 0
    
    @property
    def cpu_efficiency(self):
        """
        Calculate the CPU efficiency of this calculation.

        :param max_cpu: The number of allocated CPUs, this will be guestimated automatically if not available.
        """
        # Integrate to find the number of CPU seconds used.
        area = integrate.trapezoid(self.cpu_used, self.duration)

        # How much we should/could have used.
        total_area = (self.duration[-1] - self.duration[0]) * self.cpu_allocated * 100

        # Return as %.
        try:
            return area / total_area * 100

        except Exception:
            # Div zero
            return 0
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        duration = [0.0] * len(data['values'])
        memory_used = [0.0] * len(data['values'])
        memory_allocated = Memory(data['memory_allocated']['value'])
        memory_used_percent = [0.0] * len(data['values'])
        memory_available = [0.0] * len(data['values'])
        memory_available_percent = [0.0] * len(data['values'])
        cpu_used = [0.0] * len(data['values'])
        cpu_allocated = data['cpu_allocated']
        output_available = [0.0] * len(data['values'])
        scratch_used = [0.0] * len(data['values'])
        scratch_available = [0.0] * len(data['values'])

        for i, value in enumerate(data['values']):
            duration[i] = value['duration']['value']
            memory_used[i] = value['memory_used']['value']
            memory_used_percent[i] = value['memory_used_percent']['value']
            memory_available[i] = value['memory_available']['value']
            memory_available_percent[i] = value['memory_available_percent']['value']
            cpu_used[i] = value['cpu_used']['value']
            output_available[i] = value['output_space']['value']
            if 'scratch_used' in value:
                scratch_used[i] = value['scratch_used']['value']
            scratch_available[i] = value['scratch_space']['value']
        
        return self(
            duration = duration,
            memory_used = memory_used,
            memory_allocated = memory_allocated,
            memory_used_percent = memory_used_percent,
            memory_available = memory_available,
            memory_available_percent = memory_available_percent,
            cpu_used = cpu_used,
            cpu_allocated = cpu_allocated,
            output_available = output_available,
            scratch_used = scratch_used,
            scratch_available = scratch_available
        )

    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "cpu_allocated": self.cpu_allocated,
            "cpu_efficiency": {
                "units": "%",
                "value": float(self.cpu_efficiency),
            },
            "memory_allocated": {
                "units": "bytes",
                "value": float(self.memory_allocated)
            },
            "maximum_memory": {
                "units": "bytes",
                "value": self.max_mem,
            },
            "memory_margin": {
                "units": "bytes",
                "value": self.memory_margin
            },
            "memory_efficiency": {
                "units": "%",
                "value": float(self.memory_efficiency)
            },
            "values":[
                {
                    'duration': {
                        "units": "s",
                        "value": self.duration[i]
                    },
                    'memory_used': {
                        "units": "bytes",
                        "value": self.memory_used[i]
                    },
                    'memory_used_percent': {
                        "units": "%",
                        "value": self.memory_used_percent[i]
                    },
                    'memory_available': {
                        "units": "bytes",
                        "value": self.memory_available[i]
                    },
                    'memory_available_percent': {
                        "units": "bytes",
                        "value": self.memory_available_percent[i]
                    },
                    'cpu_used': {
                        "units": "%",
                        "value": self.cpu_used[i]
                    },
                    'output_space': {
                        "units": "bytes",
                        "value": self.output_space[i]
                    },
                    'scratch_used': {
                        "units": "bytes",
                        "value": self.scratch_used[i]
                    },
                    'scratch_space': {
                        "units": "bytes",
                        "value": self.scratch_space[i]
                    }
                } for i in range(len(self.duration))
            ]
        }

        return {
            'duration': {
                "units": "s",
                "values": self.duration.tolist()
            },
            'memory_used': {
                "units": "bytes",
                "values": self.memory_used.tolist()
            },
            'memory_used_percent': {
                "units": "%",
                "values": self.memory_used_percent.tolist()
            },
            'memory_available': {
                "units": "bytes",
                "values": self.memory_available.tolist()
            },
            'memory_available_percent': {
                "units": "bytes",
                "values": self.memory_available_percent.tolist()
            },
            'cpu_used': {
                "units": "%",
                "values": self.cpu_used.tolist()
            },
            'output_space': {
                "units": "bytes",
                "values": self.output_space.tolist()
            },
            'scratch_space': {
                "units": "bytes",
                "values": self.scratch_space.tolist()
            }
        }




"""Utilities and handy functions for reading and parsing result files."""

# General imports.
from functools import partial
from itertools import filterfalse, zip_longest
from pathlib import Path
import itertools
import shutil
from tempfile import mkdtemp
import collections
import warnings
# IMPORTANT: Do not replace multiprocessing pools with pathos, the latter is too buggy for production ATM (26-05-2023).
import multiprocessing

from configurables.misc import is_iter, is_int
from configurables.defres import Default, defres

# Digichem imports.
from digichem.exception.base import Digichem_exception
from digichem.parse.cclib import Cclib_parser
from digichem.parse.gaussian import Gaussian_parser
from digichem.parse.turbomole import Turbomole_parser
from digichem.parse.orca import Orca_parser
from digichem.result.multi import Merged
from digichem.result.result import Result_set
import digichem.log
from digichem.parse.dump import Yaml_multi_parser, Json_multi_parser

# Hidden imports.
#import cclib.io
#import cclib.parser

custom_parsing_formats = {
    "sir": Yaml_multi_parser,
    "yaml": Yaml_multi_parser,
    "sij": Json_multi_parser,
    "json": Json_multi_parser,
}

archive_formats = list(itertools.chain(*[extensions for name, extensions, desc in shutil.get_unpack_formats()]))

def find_log_files_from_hint(hint):
    """
    Find output (log) files from a given hint.
    
    :param hint: A path to a file to use as a hint to find additional log files. hint can optionally be a directory, in which case files inside this directory will be found.
    :returns: A list of found log files.
    """
    hint = Path(hint)
    
    # First, find our parent dir.
    # hint may actually be a dir.
    if hint.is_dir():
        # Look for all .log files.
        # File extensions that we recognise.
        log_types = itertools.chain(["*." + custom_format for custom_format in custom_parsing_formats], ["*.log", ]) # "*.out" disabled for now...
        parent = hint
        log_files = [found_log_file for found_log_file in itertools.chain(*[parent.glob(log_type) for log_type in log_types])]
        
        # Remove any 'digichem.log' files as we know these are not calc log files.
        # We don't actually write 'digichem.log' files anymore either (we use digichem.out instead),
        # but older versions did...
        log_files = [log_file for log_file in log_files if log_file.name not in ["digichem.log", "digichem.out", "silico.log", "silico.out", ".digichem.yaml"]]
    else:
        parent = hint.parent
        log_files = [hint]
    
    # If we have a computational style log file, look for others.
    if hint.suffix not in ["." + custom_format for custom_format in custom_parsing_formats]:
        # Try and find job files.
        # These files have names like 'job.0', 'job.1' etc, ending in 'job.last'.
        for maybe_job_file in parent.glob("job.*"):
            # We only want the numbered files.
            if is_int(".".join(maybe_job_file.name.split(".")[1:])):
                # Looks good.
                log_files.append(maybe_job_file)
        
        # Look for other files.
        for maybe_file_name in ("basis", "control", "mos", "alpha", "beta", "coord", "gradient", "aoforce", "job.last", "numforce/aoforce.out"):
            maybe_file_path = Path(parent, maybe_file_name)
            
            if maybe_file_path.exists():
                # Found it.
                log_files.append(maybe_file_path)
            
    # Make sure we only have unique log files.
    # We also now reverse our ordering, so that files given earlier by the user have priority.
    if len(log_files) > 0:
        return log_files
            
    # If we have no log files, and there's a directory called Output or Result that we can use, try again using that as the hint.
    elif hint.is_dir():
        # Try the Output dir first.
        if Path(hint, "Output").is_dir():
            log_files = find_log_files_from_hint(Path(hint, "Output"))
            
        # If we still have nothing, try the Results dir.
        if len(log_files) == 0 and Path(hint, "Results").is_dir():
            log_files = find_log_files_from_hint(Path(hint, "Results"))
            
    return log_files


def find_log_files(*hints):
    """
    Find log files from a number of given hints.
    
    Each hint should point to an existing log file to parse, or a directory containing such log files.
    Each log file given (and found) should refer to the same calculation. 
    """
    # Get a list of found log files.
    log_files = [found_log for hint in hints for found_log in find_log_files_from_hint(hint)]
    
    # Make sure we only have unique log files.
    # We also now reverse our ordering, so that files given earlier by the user have priority.
    log_files = list(reversed(list(dict.fromkeys([log_file.resolve() for log_file in log_files]))))

    return log_files
    

def class_from_log_files(*log_files, format_hint = "auto"):
    """
    Get a parser class based on some calculation log files.
    
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'cclib' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    """
    import cclib.io
    import cclib.parser
    
    if format_hint in custom_parsing_formats:
        return custom_parsing_formats[format_hint]
    
    elif format_hint == "auto" and len(log_files) > 0 and log_files[0].suffix[1:] in custom_parsing_formats:
        return custom_parsing_formats[log_files[0].suffix[1:]]
    
    if format_hint not in ["cclib", "auto"]:
        raise ValueError("Unrecognised format hint '{}'".format(format_hint))
    
    # We'll use cclib to guess the file type for us.
    try:
        log_file_type = type(cclib.io.ccopen([str(found_log_file) for found_log_file in log_files]))
        
    except Exception as e:
        if isinstance(e, FileNotFoundError):
            raise

        # cclib couldn't figure out the file type, it probably wasn't a .log file.
        raise Digichem_exception("Could not determine file type of file(s): '{}'; are you sure these are computational log files?".format(", ".join((str(log_file) for log_file in log_files)))) from e
    
    # Either get a more complex parser if we have one, or just return the base parser.
    if log_file_type == cclib.parser.gaussianparser.Gaussian:
        return Gaussian_parser
    elif log_file_type == cclib.parser.turbomoleparser.Turbomole:
        return Turbomole_parser
    elif log_file_type == cclib.parser.orcaparser.ORCA:
        return Orca_parser
    else:
        return Cclib_parser

def from_log_files(*log_files, format_hint = "auto", parser_options = {}, **auxiliary_files):
    """
    Get an output file parser of appropriate type.
    
    This is a convenience function.
    
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'log' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    """
    found_log_files = find_log_files(*log_files)

    try:
        return class_from_log_files(*found_log_files, format_hint = format_hint).from_logs(*found_log_files, hints = log_files, **parser_options, **auxiliary_files)
     
    except Exception:
        if len(found_log_files) == 0:
            raise ValueError("There are no log files at '{}'".format(log_files[0] if len(log_files) == 1 else log_files)) from None
         
        else:
            raise
    
def parse_calculation(*log_files, options, parse_all = False, format_hint = "auto", keep_archive = False, parser_options = {}, **auxiliary_files):
    """
    Parse a single calculation result.
    
    :param log_files: A number of calculation result files corresponding to the same calculation.
    :param options: A Digichem options nested dictionary containing options to control parsing.
    :param parse_all: Whether to parse all results in the given log file. If True, a list of result sets will be returned, if False, only the first result will be returned if there are multiple.
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'log' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    :param auxiliary_files: Optional auxiliary calculation files corresponding to the calculation.
    :return: A single Result_set object.
    """        
    # Handle aux files.
    # Auxiliary files are files associated with a calculation but that do not contain calculation output directly (they are not log files).
    # Often, these files are written in a program-dependent binary format, and may be used for eg, program restarting, post-calculations, or image generation.
    # For example, common aux files for Gaussian include: .chk, .fchk and .rwf.
    # Auxiliary files to associate with a log file(s) can be given by the auxiliary_files key-word argument, but this is cumbersome in some instances.
    # Alternatively, aux files can be given as normal log_files by prefixing the file name with 'aux':, where aux is the type of file.
    # For example, a chk file could be given as "chk:calc/output.chk".
    real_log_files = []
    for maybe_log_file in log_files:
        maybe_log_file = str(maybe_log_file)
        found = False
        # Loop through all known auxiliary_files:
        for file_type, aux_file_name in itertools.chain(Gaussian_parser.INPUT_FILE_TYPES.items(), Turbomole_parser.INPUT_FILE_TYPES.items(), Orca_parser.INPUT_FILE_TYPES.items()):
            code_len = len(file_type.short_code) +1
            if maybe_log_file[:code_len] == file_type.short_code + ":":
                auxiliary_files[aux_file_name] = maybe_log_file[code_len:]
            
                # Done.
                found = True
                break
        
        if not found:
            # Take care of 'log:' files (which we support in-case a normal log file happened to start with 'chk:' or similar).
            if maybe_log_file[:4] == "log:":
                real_log_files.append(maybe_log_file[4:])
                
            else:
                # Take as is.
                real_log_files.append(maybe_log_file)
            
    log_files = real_log_files
        
    # Open files for reading (handles archives for us).
    archive = open_for_parsing(*log_files, auxiliary_files = auxiliary_files)
    
    try:
        open_log_files, open_aux_files = archive.open()
        
        if parse_all:
            results = from_log_files(*open_log_files, format_hint = format_hint, parser_options = parser_options, options = options, **open_aux_files).process_all()
        
        else:       
            results = from_log_files(*open_log_files, format_hint = format_hint, parser_options = parser_options, options = options, **open_aux_files).process()
        
    finally:
        if not keep_archive:
            archive.cleanup()
            
    if keep_archive:
        # We've been asked not to close the archive, return it.
        return (results, archive)
    
    else:
        # The caller isn't interested in the archive.
        return results


def multi_parser(log_files, auxiliary_files, *, options, format_hint = "auto", keep_archive = False, parser_options = {},):
        """
        The inner function which will be called in parallel to parse files.
        """
        # If the given 'log_file' is actually already a result object, then there's nothing for us to do.
        # This is allowed to support merging calculation results between previously parsed results and new log files.
        if isinstance(log_files, Result_set):
            # Nothing we need to do.
            return [log_files]
        
        # Next, decide if this is a single log file path, or is actually a list of multiple paths.
        # Regardless of whether this is a single file or multiple, all given files should correspond to the same calculation.
        if not isinstance(log_files, str) and is_iter(log_files):
            # Multiple paths.
            logs = log_files
        
        else:
            # Single path.
            logs = (log_files,)
        
        try:    
            return parse_calculation(*logs, options = options, parse_all = True, format_hint = format_hint, keep_archive = keep_archive, parser_options = parser_options, **auxiliary_files)
            
        except Exception:
            digichem.log.get_logger().warning("Unable to parse calculation result file '{}'; skipping".format(logs[0] if len(logs) == 1 else logs), exc_info = True)
            return None

def parse_multiple_calculations(*log_files, auxiliary_files = None, options, parser_options = {}, pool = None, init_func = None, init_args = None, format_hint = "auto", processes = 1, keep_archive = False):
    """
    Parse a number of separate calculation results in parallel.
    
    If the argument 'pool' is not given, a multiprocessing.pool object will first be created using the arguments init_func and num_cpu.
    
    :param log_files: A number of calculation result files corresponding to different calculations. Each item can optionally be a list itself, to specify files from the same calculation but which are spread across multiple files.
    :param auxiliary_files: A list of dicts of aux files. The ordering of the dicts should correspond to that of log_files.
    :param options: A Digichem options nested dictionary containing options to control parsing.
    :param pool: An optional subprocessing.pool object to use for parallel parsing.
    :param init_func: An optional function to call to init each newly created process.
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'log' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    :param processes: The max number of processes to create the new pool object with; if the number of given log_files is less than processes, then len(log_files) will be used instead.
    """
    if len(log_files) == 0:
        # Give up now.
        return []
    
    if auxiliary_files is None:
        auxiliary_files = [{}] * len(log_files)
    
    if len(log_files) < processes:
        processes = len(log_files)
    
    # Sort out our pool if we need to.
    own_pool = False
    if pool is None:
        own_pool = True
        pool = multiprocessing.Pool(processes, initializer = init_func, initargs = init_args if init_args is not None else [])
    
    # Do some parsing.
    try:
        result_lists = list(
            filterfalse(lambda x: x is None,
                pool.starmap(partial(multi_parser, options = options, format_hint = format_hint, keep_archive = keep_archive, parser_options = parser_options), zip_longest(log_files, auxiliary_files, fillvalue = {}))
                #pool.map(partial(multi_parser, options = options, format_hint = format_hint, keep_archive = keep_archive), *transpose(list(zip_longest(log_files, auxiliary_files, fillvalue = {})), 2))
            )
        )
        
        if keep_archive:
            return [(result, archive) for results, archive in result_lists for result in results]
        
        else:
            return [result for result_list in result_lists for result in result_list]
    
    finally:
        # Do some cleanup if we need to.
        if own_pool:
            pool.__exit__(None, None, None)
    
def parse_and_merge_calculations(*log_files, auxiliary_files = None, options, parser_options = {}, format_hint = "auto", inner_pool = None, keep_archive = False):
    """
    Get a single result object by parsing a number of computational log files.
    
    Multiple log files can be given both from the same calculation, or from multiple different calculations.
    If multiple different calculations are given, the individually parsed results will be merged together (which may give bizarre results if the calculations are unrelated, eg if they are of different molecules).
    
    Example:
        parse_and_merge_calculations(['calc1/primary.log', 'calc1/secondary.log'], 'calc2/calc.log', 'calc3/calc.log')
    Would parse three separate calculations (calc1, calc2 and calc3), of which the first is contained in two output files (primary.log and secondary.log), merging the result sets together.
    
    :param log_files: A list of paths to computational chemistry log files to parse. If more than one file is given, each is assumed to correspond to a separate calculation in which case the parsed results will be merged together. In addition, each given 'log file' can be an iterable of log file paths, which will be considered to correspond to an individual calculation.
    :param options: A Digichem options nested dictionary containing options to control parsing.: An alignment class to use to reorientate each molecule.
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'log' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    :param auxiliary_files: A list of dictionaries of auxiliary files. The ordering of auxiliary_files should match that of log_files.
    :return: A single Result_set object (or child thereof).
    """    
    parsed_results = parse_multiple_calculations(*log_files, options = options, parser_options = parser_options, format_hint = format_hint, pool = inner_pool, auxiliary_files = auxiliary_files, keep_archive = keep_archive)
    
    # If we asked for archives as well, unpack.
    if keep_archive:
        parsed_results, archives = list(map(list, zip(*parsed_results)))
    
    # If we have more than one result, merge them together.
    if len(parsed_results) > 1:
        parsed_results = Merged.from_results(*parsed_results, options = options)
    
    elif len(parsed_results) == 0:
        parsed_results = None
    
    else:
        parsed_results = parsed_results[0]
        
    if keep_archive:
        return (parsed_results, archives)
    
    else:
        return parsed_results
            
def multi_merger_parser(log_files, auxiliary_files, *, options, parser_options = {}, format_hint = "auto" , inner_pool = None, keep_archive = False):
        """
        The inner function which will be called in parallel to parse files.
        """
        try:
            return parse_and_merge_calculations(*log_files, options = options, parser_options = parser_options, format_hint = format_hint, inner_pool = inner_pool, auxiliary_files = auxiliary_files, keep_archive = keep_archive)
            
        except Exception:
            digichem.log.get_logger().warning("Unable to parse and merge calculation results '{}'; skipping".format(", ".join([str(log_file) for log_file in log_files])), exc_info = True)
            return None

def parse_and_merge_multiple_calculations(*multiple_results, options, parser_options = {}, format_hint = "auto", init_func = None, init_args = None, processes = None, auxiliary_files = None, keep_archive = False):
    """
    Parse a number of separate calculation results in parallel, merging some or all of the results into combined result sets.
    
    :param multiple_results: A list of two dimensions, where the first dimension is a list of separate results to process, and the second dimension is a list of results that should be merged together.
    :param options: A Digichem options nested dictionary containing options to control parsing.
    :param format_hint: A hint as to the format of the given log files. Either 'auto' (to guess), 'log' (calc log file), 'sir' (digichem result file) or 'sid' (digichem database file).
    :param init_func: An optional function to call to init each newly created process.
    :param processes: The max number of processes to create the new pool object with.
    :param auxiliary_files: An optional list of lists of dicts of auxiliary files. Each item in auxiliary_files should match the corresponding log file in multiple_results.
    :return: A single Result_set object (or child thereof).
    """
    if auxiliary_files is None:
        auxiliary_files = [None] * len(multiple_results)
    
    pool = None
    # Do some parsing.
    # TODO: This parallelization isn't ideal, currently we process each group of to-be merged calcs separately, meaning processes can be wasted.
    try:
        pool = multiprocessing.Pool(processes, initializer = init_func, initargs = init_args if init_args is not None else [])
        
        result_lists = list(
            filterfalse(lambda x: x is None,
                map(partial(multi_merger_parser, options = options, parser_options = parser_options, format_hint = format_hint, inner_pool = pool, keep_archive = keep_archive), multiple_results, auxiliary_files)
            )
        )
        
        return result_lists
    
    finally:
        if pool:
            pool.__exit__(None, None, None)
    

class open_for_parsing():
    """
    A context manager for opening log files for parsing. Returns a list of pathlib.Path objects suitable for parsing.
    
    Currently, the main purpose of this context manager is to intelligently handle unpacking of archives (.zip, .tar etc) for parsing.
    """
    
    def __init__(self, *log_files, auxiliary_files = Default(None)):
        log_files = [Path(log_file).resolve() for log_file in log_files]
        
        # Check we haven't been given any duplicate log files.
        # This is just for convenience, if duplicates have been given the user has probably made a mistake.               
        duplicates = [path for path, number in collections.Counter(log_files).items() if number > 1]
        
        for duplicate in duplicates:
            warnings.warn("Ignoring duplicate log file: ".format(duplicate))
        
        # Remove duplicates but retain order.
        self.log_files = list(dict.fromkeys(log_files).keys())

        # Should we worry about duplicate aux files?
        self.has_aux_files = not isinstance(auxiliary_files, Default)
        self.auxiliary_files = {aux_type: Path(auxiliary_file).resolve() for aux_type, auxiliary_file in auxiliary_files.items()} if defres(auxiliary_files) is not None else {}
        
        # A list of tempfile.TemporaryDirectory objects that should be closed when we are finished.
        #self.temp_dirs = []

        self.archive_dirs = {}

        # Keep track of past work to prevent unpacking duplicates.
        self.done_archives = []
    
    @classmethod
    def get_archive_formats(self):
        """
        Get a list of supported archive formats.
        """
        return archive_formats
    
    @property
    def archive_formats(self):
        return archive_formats
        
    def __enter__(self):
        """
        'Open' files for reading.
        """
        return self.open()
    
    @classmethod
    def is_archive(self, path):
        """
        Determine (based on a given file extensions) whether a path points to an archive.
        """
        # We can't simply join all the extensions together, what if the file is named something like "Benzene.log.zip"
        # Also be careful for files like Benzene.tar.gz, this is .tar.gz not .gz
        for pos in range(0, len(path.suffixes)):
            if "".join(path.suffixes[pos:]) in self.get_archive_formats():
                # This is an archive format.                
                return True
        
        return False
    
    def find(self, hint):
        """
        """
        # First, open the file if it is an archive.
        open_files = []

        if hint.exists() and self.is_archive(hint):
            open_files = self.extract(hint)
        
        else:
            open_files = [hint]

        return open_files
        
    def open(self):
        """
        'Open' files for reading.
        """
        new_log_files = []
        
        # First, unpack any explicitly specified aux files.
        # We do this first because we won't extract the same file twice, and we want to make sure
        # we definitely capture the file here.
        new_aux_files = {}
        for aux_type, path in self.auxiliary_files.items():
            if self.is_archive(path):
                files = list(self.extract(path))

                # If the archive contains multiple files, complain (because we don't know which one the user wants).
                if len(files) == 0:
                    raise Digichem_exception("Cannot extract auxiliary file archive '{}'; this archive is empty".format(path))

                elif len(files) > 1:
                    raise Digichem_exception("Cannot extract auxiliary file archive '{}'; this archive contains multiple files".format(path))

                new_aux_files[aux_type] = files[0]
            
            else:
                new_aux_files[aux_type] = path
        
        # Next, build a list of files and folders to check.
        for log_file in self.log_files:
            found_child_archive = False
            
            if log_file.is_dir():
                parent_dir = log_file
                
                # If 'log_file' is a directory, check for an archive inside called 'Output.xxx'.
                for archive_format in self.archive_formats:
                    child_archive = Path(log_file, "Output" + archive_format)
                    if child_archive.exists():
                        if not found_child_archive:
                            # Found an Output dir archive, use this instead.
                            new_log_files.extend(self.find(child_archive))
                            found_child_archive = child_archive
                        
                        else:
                            # For now, only care about the first.
                            warnings.warn("Ignoring subsequent Output archive '{}'; already found '{}'".format(child_archive, found_child_archive))
                
            else:
                # The hint is not a directory, either it is a normal log file, or an archive.
                # If it is an archive, it could either be:
                #  - An archive of a single file (eg, Output.log.zip).
                #  - An archive of an Output dir.
                #  - An archive of a calculation dir (inside of which is Output).
                new_files = list(self.find(log_file))
                new_log_files.extend(new_files)
                parent_dir = log_file.parent
                
                if not all([file.is_file() for file in new_files]):

                #if "Output" in [file.name for file in new_files] or len(list(itertools.chain(*[file.glob("Output") for file in new_files]))) > 0:
                    found_child_archive = True
                
                elif self.is_archive(log_file):
                    # If the hint is a single file archive, also add the parent dir (incase not all of the files are archives).
                    new_log_files.append(log_file.parent)


            if not found_child_archive:
                # If there's not an Output.zip type file, also look for individually zipped output files.
                for sub_file in parent_dir.iterdir():
                    if self.is_archive(sub_file):
                        new_log_files.extend([found.parent for found in self.find(sub_file)])
                        
                new_log_files.extend(self.find(log_file))
        
        if self.has_aux_files:
            return new_log_files, new_aux_files
        
        else:
            return new_log_files
    
    def extract(self, file_name):
        """
        Extract an archive and return the contained log files.
        """
        file_name = file_name.resolve()
        if file_name in self.done_archives:
            digichem.log.get_logger().debug("Skipping duplicate archive '{}'".format(file_name))
            return []

        else:
            self.done_archives.append(file_name)
        
        # Get a temp dir to extract to.
        # We can't use TemporaryDirectory here, because these are auto deleted on program exit. This is not compatible with multi-processing.
        tempdir = mkdtemp()
        self.archive_dirs[file_name] = tempdir
        
        # Extract to it.
        digichem.log.get_logger().info("Extracting archive '{}'...".format(file_name))
        shutil.unpack_archive(file_name, tempdir)
        
        # Add any files/directories that were unpacked.
        return Path(tempdir).glob("*")
    
    def cleanup(self):
        """
        Perform cleanup of any open files.
        
        Alias for __exit__().
        """
        return self.__exit__()
    
    def __exit__(self, exc_type = None, exc_value = None, traceback = None):
        """
        'Close' any open files.
        """
        for tempdir in self.archive_dirs.values():
            shutil.rmtree(tempdir, ignore_errors = True)
        
        


# Objects for making cube files.

from pathlib import Path
import subprocess
import tempfile
import asyncio
import os

try:
    from pyscf.tools import cubegen

except ModuleNotFoundError:
    # No PySCF.
    cubegen = None

from digichem.exception.base import File_maker_exception
from digichem.file import File_converter
import digichem.file.types as file_types
import digichem.log
from digichem.file.base import File_maker, Dummy_file_maker
from digichem.memory import Memory
from digichem.misc.io import expand_path


def sanitize_modern_cubes(cube_file_path):
    """
    'Sanitize' or 'fix' modern cube files to be compatible with older parsers.
    
    Modern cubes can store more than one density per coordinate, but many programs don't like this.
    
    :param cube_file: Path to the cube file to modify.
    """
    cube_file_path = Path(cube_file_path)
    with open(cube_file_path, "rt") as old_cube_file, \
        tempfile.NamedTemporaryFile("wt", dir = cube_file_path.parent, delete = False) as temp_cube:
        for line_index, line in enumerate(old_cube_file):
            # Line 3 (index 2) contains this header:
            #   -11  -10.884822  -13.392441  -10.884822    1
            # The final '1' is unsupported on older software.
            if line_index == 2:
                line_split = line.split()
                if len(line_split) == 5:
                    # We have a line that needs cleaning.
                    if line_split[-1] != "1":
                        digichem.log.get_logger().warning(
                            "Cube file {} contains more than 1 density ('{}' densities found), this file may no longer be fully compliant".format(
                                cube_file_path,
                                line_split[-1]
                            )
                        )
                    
                    line = " " + " ".join(line_split[:-1]) + "\n"
            
            temp_cube.write(line)
            
        # Replace old cube with new cube.
        os.rename(temp_cube.name, cube_file_path)
    

class Fchk_to_cube(File_converter):
    """
    Class for handling file references to Gaussian cube files.
    
    We currently rely on Gaussian's cubegen facility to make cube files. There are a number of non-trivial problems with this approach:
     - Cubegen is part of the (very expensive) Gaussian package and so is not readily available on most machines used for post-processing.
     - The cubegen interface is poor, which makes effective error reporting particularly challenging.
     - Cubegen only takes .fchk files as input, while the rest of this analysis package reads the .log file.
       - .fchk files are not even generated by Gaussian automatically, they must be created from .chk files using another buggy Gaussian program: formchk
       - The requirement on these .fchk files is unnecessary as the data is also available in the .log file, but cubegen cannot read it.
    """
    
    # Text description of our input file type, used for error messages etc.
    #input_file_type = "fchk"
    input_file_type = file_types.gaussian_fchk_file
    # Text description of our output file type, used for error messages etc.
    output_file_type = file_types.gaussian_cube_file
    
    def __init__(
            self,
            *args,
            fchk_file = None,
            cubegen_type = "MO",
            orbital = "HOMO",
            npts = 0,
            cube_file = None,
            num_cpu = None,
            memory = None,
            cubegen_executable = "cubegen",
            sanitize = False,
            multithreading = None,
            **kwargs):
        """
        Constructor for Fchk_to_cube objects.
        
        See Image_maker for a full signature.
        
        :param output: The filename/path to the cube file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param fchk_file: Optional fchk_file to use to generate this cube file.
        :param cubegen_type: The type of orbital that will be included in the cube file when we make it. Possible values are MO, AMO and BMO (for MO, alpha MO and beta MO respectively).
        :param orbital: The orbital to be included in the cube file when we make it. Possible values are 'HOMO', 'LUMO' or the integer level of the desired orbital.
        :param npts: The 'npts' option of cubegen, controls how detailed the resulting file is. Common options are 0 (default), -2 ('low' quality), -3 (medium quality), -4 (very high quality).
        :param cube_file: An optional file path to an existing cube file to use. If this is given (and points to an actual file), then a new cube will not be made and this file will be used instead.
        :param num_cpu: The number of CPUs/processes for cubegen to use. If not given, the number of CPUs available will be used.
        :param memory: The amount of memory for cubegen to use.
        :param sanitize: Whether to modify the cube file to make it compatible with older software.
        """
        super().__init__(*args, input_file = fchk_file, existing_file = cube_file, **kwargs)
        self.cubegen_type = cubegen_type
        self.orbital = orbital
        self.npts = npts
        memory = memory if memory is not None else "3 GB"
        self.memory = Memory(memory)
        # Default to number of CPUs of the system.
        # If we've not been asked to run multithreaded, use 1 CPU.
        if multithreading is None or multithreading == "pool":
            self.num_cpu = 1
        
        else:
            self.num_cpu = int(num_cpu) if num_cpu is not None else os.cpu_count()
        
        self.cubegen_executable = expand_path(cubegen_executable)
        self.sanitize = sanitize
        
    @classmethod
    def from_options(self, output, *, fchk_file = None, cubegen_type = "MO", orbital = "HOMO", num_cpu = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            fchk_file = fchk_file,
            cubegen_type = cubegen_type,
            orbital = orbital,
            npts = options['render']['orbital']['cube_grid_size'].to_gaussian(),
            dont_modify = not options['render']['enable_rendering'],
            cubegen_executable = options['external']['cubegen'],
            sanitize = options['render']['safe_cubes'],
            num_cpu = num_cpu,
            multithreading = options['external']['cubegen_parallel'],
            **kwargs
        )
    
    def get_parallel(self, name = 'file', cpus = 1):
        """
        Generate the file represented by this object in a parallel context.

        get_parallel() will be called by a higher level function as an argument to ThreadPoolExecutor.map() or similar,
        to generate many files simultaneously. This is useful for slow operations (such as cube generation) that are
        difficult to parallelise individually, but easy to parallelise across multiple files.

        :param name: The file to generate.
        :param cpus: The number of CPUs this operation should use, nearly always 1.
        """
        # Temporarily set CPUs to 1 (in-case it got changed somewhere else.)
        old_cpus = self.num_cpu
        self.num_cpu = cpus
        try:
            self.get_file(name = name)
            
        finally:
            self.num_cpu = old_cpus

    @property
    def signature(self):
        """
        Get the call signature to run cubegen (as can be passed to Popen etc)
        """
        return [
            "{}".format(self.cubegen_executable),
            # Some versions of cubegen respect this num_threads arg, some don't.
            "{}".format(self.num_cpu),
            "{}={}".format(self.cubegen_type, self.orbital),
            str(self.input_file),
            str(self.output),
            str(self.npts)
        ]

    @property
    def env(self):
        """
        Get the environmental variables require to run cubegen (as can be passed to Popen etc).
        """
        return dict(
            os.environ,
            GAUSS_MEMDEF = str(self.memory),
            # Some versions of cubegen only respect this way of changing threads.
            OMP_NUM_THREADS = str(self.num_cpu)
        )

    async def make_files_async(self):
        """
        Make the files referenced by this object.
        """        
        try:
            cubegen_proc =  await asyncio.create_subprocess_exec(
                # Unless subprocess.run(), create_subprocess_exec() takes and expanded list as first arg...
                *self.signature,
                # Capture both stdout and stderr.
                # It appears that cubegen writes everything (including error messages) to stdout, but it does return meaningful exit codes.
                stdout = subprocess.PIPE,
                stderr = subprocess.STDOUT,
                env = self.env
            )
        except FileNotFoundError:
            raise File_maker_exception(self, "Could not locate cubegen executable '{}'".format(self.cubegen_executable))
        
        stdout = ""
        stderr = ""
        # Wait for completion.
        while cubegen_proc.returncode is None:
            retval = await cubegen_proc.communicate()
            stdout += retval[0].decode()
            stderr += retval[1].decode()

        # If something went wrong, dump output.
        if cubegen_proc.returncode != 0:
            # An error occured.
            # Check if the input file exists, if not this is probably what went wrong.
            message = "Cubegen did not exit successfully"
            if not Path(str(self.input_file)).exists():
                message += " (probably because the input file '{}' could not be found)".format(self.input_file)
            message += ":\n{}".format(stdout)
            raise File_maker_exception(self, message)
        else:
            # Everything appeared to go ok.
            # Dump cubegen output if we're in debug.
            digichem.log.get_logger().debug(stdout)
        
        if self.sanitize:
            sanitize_modern_cubes(self.output)

            
    def make_files(self):
        """
        Make the files referenced by this object.
        """
        try:
            cubegen_proc =  subprocess.run(
                self.signature,
                # Capture both stdout and stderr.
                # It appears that cubegen writes everything (including error messages) to stdout, but it does return meaningful exit codes.
                stdout = subprocess.PIPE,
                stderr = subprocess.STDOUT,
                universal_newlines = True,
                env = self.env
            )
        except FileNotFoundError:
            raise File_maker_exception(self, "Could not locate cubegen executable '{}'".format(self.cubegen_executable))
        
        # If something went wrong, dump output.
        if cubegen_proc.returncode != 0:
            # An error occured.
            # Check if the input file exists, if not this is probably what went wrong.
            message = "Cubegen did not exit successfully"
            if not Path(str(self.input_file)).exists():
                message += " (probably because the input file '{}' could not be found)".format(self.input_file)
            message += ":\n{}".format(cubegen_proc.stdout)
            raise File_maker_exception(self, message)
        else:
            # Everything appeared to go ok.
            # Dump cubegen output if we're in debug.
            digichem.log.get_logger().debug(cubegen_proc.stdout)
        
        if self.sanitize:
            sanitize_modern_cubes(self.output)


class Fchk_to_spin_cube(Fchk_to_cube):
    """
    A variation of the cube maker designed for making spin density cubes.
    
    Note that this class exists mainly for convenience; either class can be used to generate cubes of any supported type.
    """
    
    def __init__(self, *args, spin_density = "SCF", **kwargs):
        """
        Constructor for Fchk_to_cube objects.
        
        See Fchk_to_cube for a full signature.
        
        :param output: The filename/path to the cube file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param fchk_file: Optional fchk_file to use to generate this cube file.
        :param spin_density: The density to use; you almost certainly want SCF unless perhaps you are using MP2...
        :param npts: The 'npts' option of cubegen, controls how detailed the resulting file is. Common options are 0 (default), -2 ('low' quality), -3 (medium quality), -4 (very high quality).
        :param cube_file: An optional file path to an existing cube file to use. If this is given (and points to an actual file), then a new cube will not be made and this file will be used instead.
        """
        return super().__init__(*args, cubegen_type = "Spin", orbital = spin_density, **kwargs)
    
    @classmethod
    def from_options(self, output, *, fchk_file = None, spin_density = "SCF", num_cpu = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            fchk_file = fchk_file,
            spin_density = spin_density,
            npts = options['render']['spin']['cube_grid_size'].to_gaussian(),
            dont_modify = not options['render']['enable_rendering'],
            cubegen_executable = options['external']['cubegen'],
            sanitize = options['render']['safe_cubes'],
            num_cpu = num_cpu,
            multithreading = options['external']['cubegen_parallel'],
            **kwargs
        )

class Fchk_to_density_cube(Fchk_to_cube):
    """
    A variation of the cube maker designed for making density cubes.
    
    Note that this class exists mainly for convenience; either class can be used to generate cubes of any supported type.
    """
    
    def __init__(self, *args, density_type = "SCF", **kwargs):
        """
        Constructor for Fchk_to_cube objects.
        
        See Fchk_to_cube for a full signature.
        
        :param output: The filename/path to the cube file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param fchk_file: Optional fchk_file to use to generate this cube file.
        :param density_type: The density to use.
        :param npts: The 'npts' option of cubegen, controls how detailed the resulting file is. Common options are 0 (default), -2 ('low' quality), -3 (medium quality), -4 (very high quality).
        :param cube_file: An optional file path to an existing cube file to use. If this is given (and points to an actual file), then a new cube will not be made and this file will be used instead.
        """
        super().__init__(*args, cubegen_type = "Density", orbital = density_type, **kwargs)
        self.type = density_type
    
    @classmethod
    def from_options(self, output, *, fchk_file = None, density_type = "SCF", num_cpu = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            fchk_file = fchk_file,
            density_type = density_type,
            npts = options['render']['density']['cube_grid_size'].to_gaussian(),
            dont_modify = not options['render']['enable_rendering'],
            cubegen_executable = options['external']['cubegen'],
            sanitize = options['render']['safe_cubes'],
            num_cpu = num_cpu,
            multithreading = options['external']['cubegen_parallel'],
            **kwargs
        )

class Fchk_to_nto_cube(Fchk_to_cube):
    """
    A variation of the cube maker designed for making NTO cubes.
    
    Note that this class exists mainly for convenience; either class can be used to generate cubes of any supported type.
    """
    
    def __init__(self, *args, orbital = "HOMO", **kwargs):
        """
        Constructor for Fchk_to_cube objects.
        
        See Fchk_to_cube for a full signature.
        
        :param output: The filename/path to the cube file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param fchk_file: Optional fchk_file to use to generate this cube file.
        :param orbital: The NTO to plot.
        :param npts: The 'npts' option of cubegen, controls how detailed the resulting file is. Common options are 0 (default), -2 ('low' quality), -3 (medium quality), -4 (very high quality).
        :param cube_file: An optional file path to an existing cube file to use. If this is given (and points to an actual file), then a new cube will not be made and this file will be used instead.
        """
        super().__init__(*args, cubegen_type = "MO", orbital = orbital, **kwargs)
    
    @classmethod
    def from_options(self, output, *, fchk_file = None, orbital = "HOMO", num_cpu = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            fchk_file = fchk_file,
            orbital = orbital,
            npts = options['render']['natural_transition_orbital']['cube_grid_size'].to_gaussian(),
            dont_modify = not options['render']['enable_rendering'],
            cubegen_executable = options['external']['cubegen'],
            sanitize = options['render']['safe_cubes'],
            num_cpu = num_cpu,
            multithreading = options['external']['cubegen_parallel'],
            **kwargs
        )

class PySCF_to_cube(File_maker):
    """
    Generate cubes from a completed PySCF calculation result object.
    """

    def __init__(
            self,
            *args,
            mol,
            target,
            target_type = "density",
            npts = 80,
            cube_file = None,
            sanitize = False,
            **kwargs):
        """
        Constructor for Fchk_to_cube objects.
        
        See Image_maker for a full signature.
        
        :param output: The filename/path to the cube file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param npts: The number of points per side of the cube.
        :param cube_file: An optional file path to an existing cube file to use. If this is given (and points to an actual file), then a new cube will not be made and this file will be used instead.
        :param sanitize: Whether to modify the cube file to make it compatible with older software.
        """
        super().__init__(*args, existing_file = cube_file, **kwargs)
        self.npts = npts
        self.sanitize = sanitize
        self.target_type = target_type
        self.target = target
        self.mol = mol
        # TODO: Add some intelligence to this...
        self.type = "SCF"
        
    @classmethod
    def from_options(self, output, *, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            npts = options['render']['orbital']['cube_grid_size'].translate("points"),
            dont_modify = not options['render']['enable_rendering'],
            sanitize = options['render']['safe_cubes'],
            **kwargs
        )
    
    def check_can_make(self):
        super().check_can_make()

        if cubegen is None:
            raise File_maker_exception(self, "PySCF is not available")
    
    def make_files(self):
        """
        Make the files referenced by this object.
        """
        if self.target_type == "density":
            cubegen.density(self.mol, self.output, self.target.make_rdm1(), nx=self.npts, ny=self.npts, nz=self.npts)

        elif self.target_type == "orbital":
            cubegen.orbital(self.mol, self.output, self.target, nx=self.npts, ny=self.npts, nz=self.npts)
        
        else:
            raise ValueError("Unrecognised 'target_type': {}".format(self.target_type))
        
        if self.sanitize:
            sanitize_modern_cubes(self.output)


# Render something with Beautiful Atoms.
# To use this script, run something like:
#
# # To disabled possibly conflicting packages outside of the conda env.
# PYTHONNOUSERSITE=1
# blender -b -P batoms-renderer.py
#
# Where 'blender' is the path to the Beautiful Atoms Blender executable.


# import debugpy
# debugpy.listen(5678)
# debugpy.wait_for_client()

import addon_utils
def handle_error(exception):
    raise exception
addon_utils.enable("batoms", handle_error = handle_error, default_set=True)

import sys
import argparse
import itertools
import bpy
import yaml
import math
from pathlib import Path
import logging

import ase.io
from batoms import Batoms
from batoms.utils.butils import object_mode
from batoms.render import Render

class Digichem_render(Render):
    def set_viewport_distance_center(self, center=None, padding=0, canvas=None):
        """
        Calculate canvas and direction
        """
        batoms = self.batoms
        if padding is None:
            padding = max(batoms.size) + 0.5
        if center is None:
            center = batoms.get_center_of_geometry()
        self.center = center
        if canvas is None:
            width, height, depth = batoms.get_canvas_box(
                direction=self.viewport, padding=padding
            )
        else:
            width = canvas[0]
            height = canvas[1]
            depth = canvas[2]
        if self.distance < 0:
            self.distance = max(10, depth)

        self.update_camera(width, height, depth / 2)

        # To auto centre the camera, we need to select the molecule as well as all isosurfaces that might be present.
        # Select the molecule.
        self.batoms.obj.select_set(True)

        # Isosurfaces.
        for obj in self.batoms.coll.all_objects:
            if obj.batoms.type == "ISOSURFACE":
                obj.select_set(True)

        # Set camera as active.
        bpy.context.scene.camera = self.camera.obj

        # Manually set a focal point.
        self.camera.lens = 50
        # Auto centre.
        bpy.ops.view3d.camera_to_view_selected()

        self.update_light()

def add_molecule(
        cube_file,
        name,
        rotations = None,
        visible = True,
        isovalues = None,
        isotype = "both",
        primary_color = [1, 0.058, 0.0, 0.55],
        secondary_color = [0.1, 0.1, 0.9, 0.55],
        style = "default"
        ):
    """
    """
    rotations = [] if rotations is None else rotations
    isovalues = [] if isovalues is None else isovalues
    
    surface_settings = []
    for isovalue in isovalues:
        if isotype in ["both", "positive"]:
            surface_settings.append({'level': isovalue, 'color': primary_color, 'material_style': style})
        
        if isotype in ["both", "negative"]:
            surface_settings.append({'level': -isovalue, 'color': secondary_color, 'material_style': style})
    
    
    # Load the input data.
    cube = ase.io.read(cube_file, format="cube", read_data=True, full_output=True)
    
    # The centre of the cube is often offset, fix that by shifting the atoms.
    cube["atoms"].translate(-cube["origin"][0:3])
    
    # Get the mol object.
    mol = Batoms(name, from_ase = cube["atoms"])
    
    # Set some look and feel options.    
    
    # Hide cell boundaries.
    mol.cell.hide = True
    
    # Colour tuning.
    # Carbon to 'black'.
    new_colors = {
        "C": (0.095, 0.095, 0.095, 1),
        "B": (1.0, 0.396, 0.468, 1)
    }

    for atom, color in new_colors.items():
        try:
            mol[atom].color = color
        except AttributeError:
            pass

    # And bonds.
    for key, value in mol.bond.settings.items():
        for atom, color in new_colors.items():
            if key[0] == atom:
                value['color1'] = color

            if key[1] == "C":
                value['color2'] = color
    
    # Change molecule style.
    mol.model_style = 1

    # Slightly increase volume of all atoms.
    for atom in mol.species.keys():
        mol[atom].scale *= 1.25
    
    # Increase volume of H atoms
    mol['H'].scale = 0.75
    
    # Add volumes.
    if len(surface_settings) != 0:
        mol.volumetric_data['surface'] = cube['data']
        
        for index, settings in enumerate(surface_settings):
            mol.isosurface.settings[index+1] = settings
        
        mol.isosurface.draw()
    
    # Now move the entire molecule (isosurface and all) back to the origin.
    mol.obj.select_set(True)
    bpy.ops.object.origin_set(type="ORIGIN_GEOMETRY", center='MEDIAN')
    bpy.context.object.location = [0,0,0]
    
    # If we have any rotations, apply those.
    for axis, angle in rotations:
        # Convert to degree.
        # Not used.
        degree = angle * (180/math.pi)
        
        if axis == 0:
            axis = "x"
        
        elif axis == 1:
            axis = "y"
        
        elif axis == 2:
            axis = "z"
        
        else:
            raise ValueError("Unknown rotation axis '{}'".format(axis))
        
        #mol.rotate(degree, axis)
        
        # Taken from batoms source (collection.py: 51)
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        mol.obj.select_set(True)
        bpy.context.view_layer.objects.active = mol.obj
        bpy.ops.transform.rotate(value=angle, orient_axis=axis.upper(),
                                 center_override = (0,0,0))
    
    if not visible:
        mol.hide = True
    
    return mol

# Adapted from https://blender.stackexchange.com/questions/5898/how-can-i-create-a-cylinder-linking-two-points-with-python?newreg=f372ba9448694f5b97879a6dab963cee
def draw_primitive(start, end, radius, mesh_type, color, collection = None):
    """
    """
    dx = end[0] - start[0]
    dy = end[1] - start[1]
    dz = end[2] - start[2]  
    dist = math.sqrt(dx**2 + dy**2 + dz**2)
    
    # First draw the shape.
    if mesh_type == "cylinder":
        bpy.ops.mesh.primitive_cylinder_add(
            radius = radius, 
            depth = dist,
            vertices = 32,
            location = (dx/2 + start[0], dy/2 + start[1], dz/2 + start[2]) 
        )
    elif mesh_type == "cone":
        bpy.ops.mesh.primitive_cone_add(
            radius1 = radius,
            depth = dist,
            vertices = 32,
            location = (dx/2 + start[0], dy/2 + start[1], dz/2 + start[2])
        )
    else:
        raise ValueError("Unknown mesh type '{}'".format(mesh_type))
        
    # Get a reference to the object we just made.
    obj = bpy.context.active_object
    
    phi = math.atan2(dy, dx)
    theta = math.acos(dz/dist) 
    
    bpy.context.object.rotation_euler[1] = theta 
    bpy.context.object.rotation_euler[2] = phi
    
    # Get material
    mat = bpy.data.materials.new(name="Material")
    
    # assign to 1st material slot
    obj.active_material = mat
        
    mat.use_nodes = True
    tree = mat.node_tree
    nodes = tree.nodes
    bsdf = nodes["Principled BSDF"]
    bsdf.inputs["Base Color"].default_value = color
    bsdf.inputs["Metallic"].default_value = 0.1
    try:
        # Blener 3.x
        bsdf.inputs["Specular"].default_value = 0.2
    except KeyError:
        # Blender 4.x
        bsdf.inputs["Specular IOR Level"].default_value = 0.2
    
    bsdf.inputs["Roughness"].default_value = 0.2
    mat.diffuse_color = color
    
    # If we've been asked to, asign our new object to a given collection.
    # First unlink from any old collections
    for coll in obj.users_collection:
        # Unlink the object
        coll.objects.unlink(obj)

    # Link each object to the target collection
    collection.objects.link(obj)

    return obj

    
def draw_arrow(start, end, radius, color, split = 0.9, collection = None):
    # Decide what proportion of the total vector to dedicate to the arrow stem and head.
    dx = end[0] - start[0]
    dy = end[1] - start[1]
    dz = end[2] - start[2]
    dist = math.sqrt(dx**2 + dy**2 + dz**2)
    
    join = (dx * split + start[0], dy * split + start[1], dz * split + start[2])
    cylinder = draw_primitive(start, join, radius, "cylinder", color, collection = collection)
    cone = draw_primitive(join, end, radius*2, "cone", color, collection = collection)

    # Select the two objects and join them together.
    bpy.ops.object.select_all(action='DESELECT')
    cylinder.select_set(True)
    cone.select_set(True)
    bpy.ops.object.join()

    arrow = cone

    # Set the origin of the new combined object to the origin of the arrow.
    bpy.context.scene.cursor.location = start
    bpy.ops.object.origin_set(type='ORIGIN_CURSOR')

    return arrow



def main():
    parser = argparse.ArgumentParser(
        prog='Beautiful Atoms Renderer',
        description='Render images with BAtoms')
    
    parser.add_argument("cube_file", help = "Path to the cube file to read")
    parser.add_argument("output", help = "File to write to", nargs="?", default = None)
    parser.add_argument("--second_cube", help = "Optional second cube file to read additional isosurface data from", default = None)
    parser.add_argument("--isovalues", help = "List of isovalues to render", nargs = "*", type = float, default = [])
    parser.add_argument("--isotype", help = "Whether to render positive, negative or both isosurfaces for each isovalue", choices = ["positive", "negative", "both"], default = "both")
    parser.add_argument("--isocolor", help = "The colouring method to use for isosurfaces", choices = ["sign", "cube"], default = "sign")
    parser.add_argument("--primary-color", help = "RGBA for one of the colors to use for isosurfaces", type = float, nargs = 4, default = [0.1, 0.1, 0.9, 0.5])
    parser.add_argument("--secondary-color", help = "RGBA for the other color to use for isosurfaces", type = float, nargs = 4, default = [1, 0.058, 0.0, 0.5])
    parser.add_argument("--style", help = "Material style for isosurfaces", choices = ('default', 'metallic', 'plastic', 'ceramic', 'mirror'), default = "ceramic")
    parser.add_argument("--cpus", help = "Number of parallel CPUs to use for rendering", type = int, default = 1)
    parser.add_argument("--use-gpu", help = "Whether to enable GPU rendering", action = "store_true")
    parser.add_argument("--orientation", help = "The orientation to render from, as x, y, z values", nargs = 3, type = float, default = [0, 0, 0])
    parser.add_argument("--resolution", help = "The output resolution in px", type = int, default = 1024)
    parser.add_argument("--render-samples", help = "The maximum number of render samples, more generally results in higher quality but longer render times", type = int, default = 64)# default = 256)
    parser.add_argument("--rotations", help = "A list of rotations (in JSON) to rotate the molecule to a given alignment. The first item in each list item is the axis to rotate about (0=x, 1=y, 2=z), the second is the angle to rotate by (in radians)", nargs = "*", default = [])
    parser.add_argument("--dipoles", help = "Draw dipoles from a list of the following data (in JSON): 0) start coord, 1) end coord, 2) RGBA color information", nargs = "*", default = [])
    parser.add_argument("--alpha", help = "Override the opacity value for all molecule objects (but not dipoles) to  1- this value, useful for showing dipole arrows more clearly", default = None, type = float)
    parser.add_argument("--perspective", help = "The perspective mode, either orthographic or perspective", default = "perspective", choices = ["perspective", "orthographic"])
    parser.add_argument("--padding", help = "Padding", type = float, default = 1.0)
    parser.add_argument("--multi", help = "Render multiple images, each of a different angle of the scene. Each argument should consist of 6 parts, the x y z position, the resolution, the samples, and the filename (which is appended to 'output')", nargs = 6, default =[], action="append")
    # Both blender and python share the same command line arguments.
    # They are separated by double dash ('--'), everything before is for blender,
    # everything afterwards is for python (except for the first argument, wich is
    # the program name, which is for both).
    if "--" in sys.argv:
        python_argv = sys.argv[sys.argv.index("--") +1:]
    else:
        python_argv = []
    
    args = parser.parse_args(python_argv)
    
    # Batoms or blender will silently set the extension to png if it's not already.
    # This is surprising, so stop now before that happens.
    if args.output is not None and Path(args.output).suffix.lower() != ".png":
        raise ValueError("Output location must have a .png extension")
    
    if args.rotations is not None:
        rotations = [yaml.safe_load(rotation) for rotation in args.rotations]

    if args.multi != []:
        if args.orientation != [0, 0, 0]:
            raise ValueError("You cannot set both --orientation and --multi!")

        if args.resolution != 1024:
            raise ValueError("You cannot set both --resolution and --multi!")
        
        if args.output is not None:
            raise ValueError("You cannot set both 'output' and --multi!")
    
    # Remove the starting cube object.
    bpy.ops.object.select_all(action='SELECT')
    bpy.ops.object.delete()

    # Load the input data.
    mol = add_molecule(
        args.cube_file,
        name = "first_mol",
        visible = True,
        rotations = rotations,
        isovalues = args.isovalues,
        isotype = args.isotype,
        primary_color = args.primary_color,
        secondary_color = args.secondary_color if args.isocolor == "sign" else args.primary_color,
        style = args.style
    )
    
    # Uncomment to show atom labels.
    # Needs some tweaking to appear in render (viewport only by default).
    #mol.show_label = 'species'
    mol2 = None
    
    # If we have a second cube, add that too.
    if args.second_cube is not None:
        mol2 = add_molecule(
            args.second_cube,
            name = "second_mol",
            visible = False,
            rotations = rotations,
            isovalues = args.isovalues,
            isotype = args.isotype,
            primary_color = args.primary_color if args.isocolor == "sign" else args.secondary_color,
            secondary_color = args.secondary_color,
            style = args.style
        )
    
    if args.alpha:
        # Set all materials transparent.
        for material in bpy.data.materials:
            try:
                material.node_tree.nodes['Principled BSDF'].inputs['Alpha'].default_value = (1 - args.alpha)
            except Exception as e:
                pass
        
    
    # Draw any dipoles.
    arrows = []
    if args.dipoles is not None:
        
        dipoles = [yaml.safe_load(dipole) for dipole in args.dipoles]
        for start_coord, end_coord, rgba in dipoles:
            arrows.append(draw_arrow(start_coord, end_coord, 0.1, rgba, collection = mol.coll))
        
    
    # Setup rendering settings.
#     mol.render.engine = 'workbench'
#     mol.render.engine = 'eevee'
    mol.render.engine = 'cycles'
    # Set up cycles for good quality rendering.
    # Prevents early end to rendering (forces us to use the actual number of samples).
    bpy.context.scene.cycles.use_adaptive_sampling = False
    # Post-processing to remove noise, works well for coloured backgrounds, useless for transparency.
    bpy.context.scene.cycles.use_denoising = True
    # Ray-tracing options
    bpy.context.scene.cycles.max_bounces = 48
    bpy.context.scene.cycles.transparent_max_bounces = 24
    if args.use_gpu:
        bpy.context.scene.cycles.device = "GPU"
    
    # Use maximum compression.
    bpy.context.scene.render.image_settings.compression = 1000
    

    # Change light intensity.
    mol.render.lights["Default"].direction = [0.1, 0.1, 1]
    mol.render.lights["Default"].obj.data.node_tree.nodes["Emission"].inputs[1].default_value = 0.2
    mol.render.lights["Default"].obj.data.angle = 0.174533

    # Add a second light for depth.
    mol.render.lights.add("Accent1", direction = [1,0.5,0.75])
    mol.render.lights.add("Accent2", direction = [0.5,1,0.75])

    mol.render.lights["Accent1"].obj.data.angle = 0.0872665
    mol.render.lights["Accent1"].obj.data.node_tree.nodes["Emission"].inputs[1].default_value = 0.25
    mol.render.lights["Accent2"].obj.data.angle = 0.0872665
    mol.render.lights["Accent2"].obj.data.node_tree.nodes["Emission"].inputs[1].default_value = 0.25

    # bpy.data.lights["batoms_light_Default"].node_tree.nodes["Emission"].inputs[1].default_value = 0.45
    # bpy.data.lights["batoms_light_Default"].angle
    #mol.render.lights["Default"].energy=10
    
    # Change view mode.
    if args.perspective == "perspective":
        mol.render.camera.type = "PERSP"
    
    else:
        mol.render.camera.type = "ORTHO"
    
    # Enable to add an outline.
    #bpy.context.scene.render.use_freestyle = True
    
    # We have plenty of memory to play with, use one tile.
    bpy.context.scene.cycles.tile_x = args.resolution
    bpy.context.scene.cycles.tile_y = args.resolution
    bpy.context.scene.cycles.tile_size = args.resolution

    # Performance options.
    bpy.context.scene.render.threads_mode = 'FIXED'
    bpy.context.scene.render.threads = args.cpus
    
    # Set our custom renderer so we can modify zoom etc.
    mol.render = Digichem_render()

    # We have two ways we can change which angle we render from.
    # 1) the viewport keyword arg (which places the camera in a certain location).
    # 2) rotate the molecule.
    #
    # We use option 2, because this gives us more control.

    # Work out how many angles we're rendering from.
    if args.multi == []:
        # Just one.
        targets = [[args.orientation[0], args.orientation[1], args.orientation[2], args.resolution, args.render_samples, args.output]]
    
    else:
        # More than one.
        targets = args.multi

    for x, y, z, resolution, samples, full_file_name in targets:
        # Add args.output and mini_file_name together (useful for --multi).
        orientation = (float(x), float(y), float(z))

        mol.render.resolution = [resolution, resolution]
        # Quality control, more = better and slower.
        bpy.context.scene.cycles.samples = int(samples)
        mol.obj.delta_rotation_euler = orientation
        
        if mol2 is not None:
            mol2.obj.delta_rotation_euler = orientation

        for arrow in arrows:
            arrow.delta_rotation_euler = orientation

        mol.get_image(viewport = [0,0,1], output = full_file_name, padding = args.padding)
    
    return 0
    
# If we've been invoked as a program, call main().    
if __name__ == '__main__':
    try:
        sys.exit(main())
    
    except Exception as e:
        logging.error("Erro", exc_info = True)
        sys.exit(1)


from deepmerge import Merger

from digichem.config.base import Digichem_options
from digichem.config.parse import Config_file_parser, Config_parser
from digichem.config.locations import master_config_path, system_config_location, user_config_location

from digichem.log import get_logger


# The main digichem options object.
# When running as a program, this will be merged with run-time options.
_options = None

def get_config(
        extra_config_files = None,
        extra_config_strings = None,
        cls = Digichem_options,
        clear_cache = False,
        sources = (
            master_config_path,
            system_config_location,
            user_config_location
        )
    ):
        """
        Get a Digichem options object.
        
        The returned object will take options from three sources:
        1) Any config files found in the default locations (see digichem.config.locations).
        2) Any additional config files specified as an argument.
        3) Any additional config strings specified as an argument.
        
        IMPORTANT: The returned options object will be merged with any additional options without prior copying.
        Hence the object returned by this function will be the same as the options attribute of this module.
        
        :param extra_config_files: An iterable of additional file paths to read from.
        :param extra_config_strings: An iterable of additional config options to parse.
        :param cls: The type of object to return.
        :param clear_cache: If True, and previously cached options will be discarded.
        :param sources: An iterable of file locations to read from.
        :return: A Digichem_options object (a fancy dict).
        """
        global _options
        if clear_cache:
            _options = None

        if _options is not None and extra_config_files is None and extra_config_strings is None:
            # Config has already been loaded (and we have nothing new to add).
            # Return the config object.
            return _options

        # Either this is the first time we've loaded the config (cache miss)
        # or we've been given extra options to add in.
        log_level = get_logger().level
        get_logger().setLevel("DEBUG")
        
        # First, load options if not already done so.
        if _options is None:
            # Load config options from given sources.
            # These objects are simple dicts.
            config = Config_file_parser(sources[0]).load(True)
            for source in sources[1:]:
                merge_dict(Config_file_parser(source).load(True), config)
                #  config.merge(Config_file_parser(source).load(True))
            
            # No need to validate here, we're going to do it later anyway.
            _options = cls(validate_now = False, **config)
        
        if extra_config_files is None:
            extra_config_files = []
            
        if extra_config_strings is None:
            extra_config_strings = []
        
        # Load any additional config files.
        for extra_config_file in extra_config_files:
            _options.deep_merge(Config_file_parser(extra_config_file).load())
            
        # Then load any additional config strings.
        for extra_config_string in extra_config_strings:
            _options.deep_merge(Config_parser(extra_config_string).load())
        
        # Check everything is valid.
        _options.validate()
        
        # Only restore log level if there was no problem.
        get_logger().setLevel(log_level)

        # And return.
        return _options

def merge_dict(new, old):
        """
        Recursively merge two dictionaries
         
        Any keys specified in new will overwrite those specified in old.
         
        :param new: A new dictionary to merge into the old.
        :param old: An old dictionary to be overwritten by new.
        """
        # Taken from deepmerge docs: https://deepmerge.readthedocs.io/en/latest/
        merger = Merger(
            # pass in a list of tuple, with the
            # strategies you are looking to apply
            # to each type.
            [
                (list, ["override"]),
                (dict, ["merge"]),
                (set, ["union"])
            ],
            # next, choose the fallback strategies,
            # applied to all other types:
            ["override"],
            # finally, choose the strategies in
            # the case where the types conflict:
            ["override"]
        )
        return merger.merge(old, new)


"""
Contains classes for representing the ground state of a system.

This file is closely related to the excited_states.py file.
These definitions may change or move.
"""

from digichem.result.excited_state import Energy_state


class Ground_state(Energy_state):
    """
    Class for representing the ground state of a system/molecule.
    """
    
    def __init__(self, charge, multiplicity, energy):
        """
        Constructor for ground state objects.
        
        :param charge: The charge of the ground states as a signed integer.
        :param multiplicity: The multiplicity as an integer of float.
        :param energy: The absolute energy of the GS in eV.
        """
        # The 'level' refers to the ordering of energy states, the ground state is always the lowest so have level 0.
        # Similarly, the multiplicity_level refers to the ordering of energy states that share the same multiplicity. Regardless of what multiplicity our GS has, we will be the lowest by definition.
        # Energy is our energy above the GS, so that will always be 0.
        super().__init__(level = 0, multiplicity = multiplicity, multiplicity_level = 0, energy = 0.0)
        self.charge = charge
        # Also save our absolute energy.
        self.absolute_energy = energy
        
    def __eq__(self, other):
        """
        Is this ground state object equal to another?
        """
        return self.charge == other.charge and self.multiplicity == other.multiplicity and self.energy == other.energy
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        parent_dict = super()._dump_(digichem_options, all)
        return {
            "index": parent_dict['index'],
            "symbol": parent_dict['symbol'],
            "charge": self.charge,
            "multiplicity": parent_dict['multiplicity'],
            "multiplicity_index": parent_dict['multiplicity_index'],
            # The 'energy' field from Energy_state is an excited state energy, relative to the ground state (so is always 0).
            # Replace with total energy.
            "energy": {
                "value": float(self.absolute_energy),
                "units": "eV"
            }
        }
    
    @classmethod
    def from_parser(self, parser):
        """
        Create a Ground_state object from an output file parser.
        
        As different energy types can be present in a single calculation, the following (arbitrary?) precedence is used:
        1) Coupled-cluster energy.
        2) The highest level Moller-Plesset energy.
        3) SCF energy.
        
        :param parser: An output file parser.
        """
        return self.from_energies(parser.results.metadata.charge, parser.results.metadata.multiplicity, parser.results.energies)
    
    @classmethod
    def from_dump(self, data, result_set, options):
        return self.from_energies(data['charge'], data['multiplicity'], result_set.energies)
    
    @classmethod
    def from_energies(self, charge, multiplicity, energies):
        """
        Create a ground state object selecting from various lists of energies.
        
        :param charge: Charge of the ground state.
        :param multiplicity: Multiplicity of the ground state.
        :param energies: List of energies.
        """
        if len(energies.cc) > 0:
            energy = energies.cc.final
        elif len(energies.mp) > 0:
            energy = energies.mp.final
        elif len(energies.scf) > 0:
            energy = energies.scf.final
        else:
            # We have no energies that we can understand. Give up.
            return None
        # Return our constructed object.
        return self(charge, multiplicity, energy)


from matplotlib import ticker
from matplotlib.transforms import Bbox
import matplotlib.figure
import math

from digichem.image import Image_maker

class Graph_image_maker(Image_maker):
    """
    A class used for creating graph images.
    
    Most of the heavy lifting here is achieved by the matplotlib library.
    """
    
    def __init__(self, *args, max_width = None, max_height = None, **kwargs):
        """
        Constructor for Graph_image_maker objects.
        
        :param max_width: The maximum image width in pixels, None for no limit.
        :param max_height: The maximum image height in pixels, None for no limit.
        """
        super().__init__(*args, **kwargs)
        
        # Set our output dpi (higher values = bigger image/quality).
        self.output_dpi = 130
        
        # The width and height (in inch) that we initially create our figure with. Note that the true figure size may change from this value.
        self.init_image_width = 5
        self.init_image_height = 5
        
        # If true, the label and tick labels will be hidden for that axis.
        self.hide_x = False
        self.hide_y = False
        
        # The label to use for the x and y axes.
        self.x_label = ""
        self.y_label = ""
        
        # The matplotlib figure object that represents our graph.
        self.figure = None
        
        # The plotting area within our figure.
        self.axes = None
        
        # Image clamping.
        self.max_width = max_width
        self.max_height = max_height
        
    def make_files(self):
        """
        Make the image described by this object.
        """
        # Change font and other defaults.
        #pyplot.rcParams.update({'font.size': 14, 'font.family': 'DejaVu Sans', 'axes.linewidth': 2})
        matplotlib.rcParams.update({'font.size': 14, 'font.family': 'DejaVu Sans', 'axes.linewidth': 2})
                
        # Make our (empty) plot.
        # pyplot uses a nasty interface, will switch to OO at some point.
        #pyplot.figure(figsize = (self.init_image_width, self.init_image_height), tight_layout = True)
        #pyplot.figure(figsize = (self.init_image_width, self.init_image_height), tight_layout = False)
        
        # Plot our graph.
        self.make_graph()
        
        # And save to file.
        self.figure.savefig(self.output, dpi = self.output_dpi)
        #pyplot.savefig(self.output, dpi = self.output_dpi)
        
    def make_graph(self):
        """
        Make the graph data described by this object.
        
        The matplotlib figure is available at self.figure (and is also returned by this method for convenience). The graph is not saved to file.
        
        :return: A reference to the created matplotlib Figure object.
        """
        # First make our figure object, which is the top-level matplotlib container.
        self.figure = matplotlib.figure.Figure(figsize = (self.init_image_width, self.init_image_height))
        
        # Also get our main axes (all the graph types we support so far contain a single plot/axes).
        self.axes = self.figure.add_subplot(111)
        
        # Call plot(), which is defined by sub-classes to actually draw the graph.
        self.plot()
        
        # Call adjust_axes(), which does what it suggests (also adds axis labels etc.)
        self.adjust_axes()
        
        # Finally, clamp our image size if we're going to be too big.
        if self.max_width is not None:
            self.clamp_dimension(0, self.max_width)
        if self.max_height is not None:
            self.clamp_dimension(1, self.max_height)
        
        # Return a reference to the figure (for convenience).
        return self.figure
        
    def plot(self):
        """
        Main workhorse of the make_files method. Inheriting classes should write their own implementation.
        
        This is called automatically by make_graph().
        """
        pass
    
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        """ 
        if self.hide_x:
            self.x_label = "Arbitrary units"
            self.axes.xaxis.set_ticks([])
        if self.hide_y:
            self.y_label = "Arbitrary units"
            self.axes.yaxis.set_ticks([])
        
        # Add our axes labels.
        if not self.x_label == "": 
            self.axes.set_xlabel(self.x_label, labelpad = 8, fontsize = 16, weight = 'bold')
        if not self.y_label == "":
            self.axes.set_ylabel(self.y_label, labelpad = 8, fontsize = 16, weight = 'bold')
    
    def clamp_dimension(self, dim, maximum):
        """
        Ensure the image does not exceed a certain number of pixels in a given dimension.
        
        :param dim: The dimension to clamp; 0 for x, 1 for y.
        :param maximum: The maximum allowed pixels.
        """
        # Check to see if we've exceeded our max size.
        fig_size = self.figure.get_size_inches()
        if (fig_size[dim] * self.output_dpi) > maximum:
            # We're too big, reduce to max.
            fig_size[dim] = maximum / self.output_dpi
            self.figure.set_size_inches(fig_size)
                        
            # Update layout. Calling this twice is necessary for some reason loool
            self.figure.tight_layout()
            self.figure.tight_layout()
    
    def constant_scale(self, axis, inch_per_axis):
        """
        Adjust the size of our figure so a given axis has a constant scale.
        
        You should adjust the corresponding axis' limits before calling this method.
        
        :param axis: The axis to adjust, 0 for the x axis, 1 for the y axis.
        :param inch_per_axis: The number of inches per unit of the given axis to adjust to. 
        """
        
        # First determine how much space we need for our margins.
        axes_position = self.axes.get_position(False).get_points()
        figure_size = self.figure.get_size_inches()
        # Axes_position is in fractions, so multiply by our current figsize.
        start_margin = axes_position[0][axis] * figure_size[axis]
        end_margin = (1 - axes_position[1][axis]) * figure_size[axis]
        
        # Next determine how much space we need for our actual data.
        axes_lim = (self.axes.get_xlim(), self.axes.get_ylim())
        axis_min = axes_lim[axis][0]
        axis_max = axes_lim[axis][1]
        #axis_min = pyplot.axis()[axis *2]
        #axis_max = pyplot.axis()[axis *2 +1]
        
        # We take the absolute because min can actually be bigger than max if our axes are inverted.
        axis_length = math.fabs(axis_max - axis_min)
        axis_size = axis_length * inch_per_axis
        
        # Update size.
        figure_size[axis] = start_margin + end_margin + axis_size
        self.figure.set_size_inches(figure_size)
        # And update the position of our axes.
        axes_position[0][axis] = start_margin / figure_size[axis]
        axes_position[1][axis] = 1 - (end_margin / figure_size[axis])
        self.axes.set_position(Bbox(axes_position))
        
    
class OneD_graph_image_maker(Graph_image_maker):
    """
    Classes for creating '1D' graphs using matplotlib's event plot.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        # Options that we'll pass to matplotlib's event plot function.
        self.plot_options = {
            "orientation": "vertical",
            "lineoffsets": 0.75,
            "linelengths": 1,
            "linewidths": None,
            "colors": None
        }
        
        # Set our output width and height.
        self.init_image_width = 4.1
        self.init_image_height = 5.6
        
    def plot(self):
        """
        Make the image described by this object.
        """
        # Plot our 1D data.
        self.plot_lines()
        
        # Annotate with data labels.
        self.plot_labels()
            
    def plot_lines(self):
        """
        Plot the lines that make up our graph.
        
        This default implementation does nothing, inheriting classes should write their own.
        """
        pass
    
    def plot_labels(self):
        """
        Plot annotation labels that explain the data in out graph.
        
        This default implementation does nothing, inheriting classes should write their own.
        """
        pass
        
class Convergence_graph_maker(Graph_image_maker):
    """
    A class for creating graphs that show the convergence of energy in optimisation calculations.
    """
    
    def __init__(self, output, energies, **kwargs):
        """
        Constructor for graph makers.
        
        :param energies: A list of energies to plot (probably an Energy_list object).
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        """
        super().__init__(output, **kwargs)
        self.energies = energies
        
        # Our output width and height (in inch).
        self.init_image_width = 7
        self.init_image_height = 5.265
        
        # Axis titles.
        self.x_label = "Step number"
        self.y_label = "Energy /eV"
        
    @classmethod
    def from_options(self, output, *, energies, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            energies = energies,
            **kwargs
        )
    
    def plot(self):
        """
        Plot the contents of our graph.
        """
        # Plot a simple 'scatter' style graph.
        self.axes.plot(range(1, len(self.energies) +1), self.energies)    
            
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        """
        super().adjust_axes()
        
        # Axes range, the format is [xmin, xmax, ymin, ymax].
        # We'll add a bit on to the y axis so we can see better.
        y_diff = max(self.energies) - min(self.energies)
        y_fudge = y_diff * 0.05
        # Also the x.
        x_diff = len(self.energies) - 1
        x_fudge = x_diff * 0.05
        
        self.axes.set_xlim(1 - x_fudge, len(self.energies) + x_fudge)
        self.axes.set_ylim(min(self.energies) - y_fudge, max(self.energies) + y_fudge)
        
        # Change spacing.
        self.axes.xaxis.set_major_locator(ticker.MaxNLocator(nbins = '12', steps = [1, 2, 2.5, 5, 10], integer = True))
        
        # Use matplotlib to automatically layout our graph.
        self.figure.tight_layout()
        
        


import pytest

from digichem.exception.uncatchable import Submission_paused, Signal_caught, Uncatchable_exception

@pytest.mark.parametrize("exception", [
    Submission_paused(None), Signal_caught(None, None)
], ids=["Submission_paused", "Signal_caught"])
def test_uncatachable(exception):
    """Can we only catch uncatachables if we really try?"""
    with pytest.raises(Uncatchable_exception):
        try:
            raise exception
        
        except Exception:
            # This should get ignored.
            pass



"""Tests for functioning of oprattle"""
from pathlib import Path

import pytest

from digichem.file.prattle import Oprattle_formats, Openprattle_converter
from digichem.test.util import ethane_xyz, benzene_cdx

@pytest.mark.parametrize("readwrite", ["read", "write"])
def test_formats(readwrite):
    if readwrite == "read":
        assert len(Oprattle_formats().read()) > 0
    
    else:
        assert len(Oprattle_formats().write()) > 0


@pytest.mark.parametrize("file_path", [
    ethane_xyz,
    benzene_cdx
])
def test_from_path(file_path):
    """
    Can we convert a file found on the filesystem?
    """
    Openprattle_converter(input_file_path = file_path).convert("com")


@pytest.mark.parametrize("file_path, mode", [
    (ethane_xyz, "r"),
    (benzene_cdx, "rb")
])
def test_from_file(file_path, mode):
    """
    Can we convert from an open file?
    """
    with open(file_path, mode) as input_file:
        Openprattle_converter(input_file = input_file, input_file_type = Path(file_path).suffix[1:]).convert("com")


@pytest.mark.parametrize("file_path, mode", [
    (ethane_xyz, "r"),
    (benzene_cdx, "rb")
])
def test_from_buffer(file_path, mode):
    """
    Can we convert from an open file?
    """
    with open(file_path, mode) as input_file:
        data = input_file.read()
        Openprattle_converter(input_file_buffer = data, input_file_type = Path(file_path).suffix[1:]).convert("com")


@pytest.mark.parametrize("file_path", [
    ethane_xyz,
    benzene_cdx
])
def test_gen3d(file_path):
    first = Openprattle_converter(input_file_path = file_path).convert("xyz", gen3D = False)

    same = Openprattle_converter(input_file_path = file_path).convert("xyz", gen3D = False)
    different = Openprattle_converter(input_file_path = file_path).convert("xyz", gen3D = True)

    assert first == same
    assert first != different


# Classes for representing MOs.

# General imports.
from itertools import filterfalse, zip_longest, chain
import warnings
import math

from digichem.exception import Result_unavailable_error
from digichem.result import Result_container
from digichem.result import Result_object
from digichem.result import Floatable_mixin


class Molecular_orbital_list(Result_container):
    """
    Class for representing a group of MOs.
    """
    
    # A warning issued when attempting to merge non-equivalent orbital lists.
    MERGE_WARNING = "Attempting to merge lists of orbitals that are not identical; non-equivalent orbitals will be ignored"
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    @property
    def HOMO_energy(self):
        """
        Get the energy of the highest occupied orbital in this list.
        
        Use get_orbital(HOMO_difference = 0) to retrieve the HOMO as an object.
        
        :raises Result_unavailable_error: If the HOMO is not available.
        :return The orbital energy (in eV).
        """
        return self.get_orbital(HOMO_difference = 0).energy
    
    @property
    def LUMO_energy(self):
        """
        Get the energy of the lowest unoccupied orbital in this list.
        
        Use get_orbital(HOMO_difference = 1) to retrieve the LUMO as an object.
        
        :raises Result_unavailable_error: If the LUMO is not available.
        :return The orbital energy (in eV).
        """
        return self.get_orbital(HOMO_difference = 1).energy
        
    @property
    def HOMO_LUMO_energy(self):
        """
        Get ΔE HOMO-LUMO; the energy difference between the HOMO and LUMO.
        
        Depending on how the orbitals are occupied, it might not make sense to calculate the HOMO-LUMO energy.
        
        :raises Result_unavailable_error: If the HOMO or LUMO is not available.
        :return: The HOMO-LUMO energy gap (in  eV).
        """
        # Get out orbitals.
        HOMO = self.get_orbital(HOMO_difference = 0)
        LUMO = self.get_orbital(HOMO_difference = 1)
        # Return the difference.
        return float(LUMO) - float(HOMO)
    
    @property
    def occupied(self):
        """
        Return a new Molecular_orbital_list containing only the occupied orbitals of this list.
        """
        return type(self)([orbital for orbital in self if orbital.is_occupied])
    
    @property
    def virtual(self):
        """
        Return a new Molecular_orbital_list containing only the virtual (unoccupied) orbitals of this list.
        """
        return type(self)([orbital for orbital in self if not orbital.is_occupied])

    @property
    def spin_type(self):
        """
        Get the spin type (alpha, beta etc.) of the orbitals in this list.
        
        :raises Result_unavailable_error: If there are no orbitals in this list.
        :return: The spin type, one of either 'alpha' or 'beta' for unrestricted calcs, 'none' for restricted calcs, or 'mixed' if multiple spin types are present in this list.
        """
        # Unique spin types in our list.
        spin_types = list(set([orbital.spin_type for orbital in self]))
        
        if len(spin_types) == 1:
            return spin_types[0]
        elif len(spin_types) > 1:
            return "mixed"
        else:
            raise Result_unavailable_error("Orbital Spin", "There are no orbitals")
    
    def get_orbital(self, criteria = None, *, label = None, HOMO_difference = None, level = None):
        """
        Retrieve an orbital based on some property.
        
        Only one of the criteria should be specified.
        
        :deprecated: Use find() instead.
        :raises Result_unavailable_error: If the requested MO could not be found.
        :param criteria: A string describing the orbital to get. The meaning of criteria is determined automatically based on its content. If criteria begins with '+' or '-' then it is used as a HOMO_difference. Otherwise, if criteria is a valid integer, then it is used as a level. Otherwise, criteria is assumed to be an orbital label.
        :param label: The label of the orbital to get.
        :param HOMO_difference: The distance from the HOMO of the orbital to get.
        :param level: The level of the orbital to get.
        :return: The Molecular_orbital object.
        """
        warnings.warn("get_orbital is deprecated, use find() instead", DeprecationWarning)
        return self.find(criteria, label = label, HOMO_difference = HOMO_difference, level = level)
    
    def find(self, criteria = None, *, label = None, HOMO_difference = None, level = None):
        """
        Retrieve an orbital based on some property.
        
        Only one of the criteria should be specified.
        
        :raises Result_unavailable_error: If the requested MO could not be found.
        :param criteria: A string describing the orbital to get. The meaning of criteria is determined automatically based on its content. If criteria begins with '+' or '-' then it is used as a HOMO_difference. Otherwise, if criteria is a valid integer, then it is used as a level. Otherwise, criteria is assumed to be an orbital label.
        :param label: The label of the orbital to get.
        :param HOMO_difference: The distance from the HOMO of the orbital to get.
        :param level: The level of the orbital to get.
        :return: The Molecular_orbital object.
        """
        # Use the search() method to do our work for us.
        return self.search(criteria, label = label, HOMO_difference = HOMO_difference, level = level, allow_empty = False)[0]
    
    def search(self, criteria = None, *, label = None, HOMO_difference = None, level = None, allow_empty = True):
        """
        Attempt to retrieve a number of orbitals based on some property.
        
        Only one of the criteria should be specified.
        
        :param criteria: A string describing the orbitals to get. The meaning of criteria is determined automatically based on its content. If criteria begins with '+' or '-' then it is used as a HOMO_difference. Otherwise, if criteria is a valid integer, then it is used as a level. Otherwise, criteria is assumed to be an orbital label.
        :param label: The label of the orbitals to get.
        :param HOMO_difference: The distance from the HOMO of the orbitals to get.
        :param level: The level of the orbitals to get.
        :param allow_empty: If False and no matching orbitals can be found, raise a Result_unavailable_error.
        :return: A (possibly empty) Molecular_orbital_list object.
        """
        # If we've been given a generic criteria, decide what it is actually talking about.        
        if criteria is not None:
            try:                
                # If our string start with a sign (+ or -) then it's a HOMO_difference.
                if criteria[:1] == "+" or criteria[:1] == "-":
                    HOMO_difference = int(criteria)
                # If its and integer (or a string that looks like one), then its a level.
                elif criteria.isdigit() or isinstance(criteria, int):
                    level = int(criteria)
                # Otherwise we assume it a label.
                else:
                    label = criteria
                
            except Exception:
                # We couldn't parse criteria, get upset.
                raise ValueError("Unable to understand given search criteria '{}'".format(criteria))
        
        # Now get our filter func.
        if label is not None:
            filter_func = lambda mo: mo.label != label
        elif HOMO_difference is not None:
            filter_func = lambda mo: mo.HOMO_difference != HOMO_difference
        elif level is not None:
            filter_func = lambda mo: mo.level != level
        else:
            raise ValueError("Missing criteria to search by; specify one of 'criteria', 'label', 'HOMO_difference' or 'level'")
        
        # Now search.
        found = type(self)(filterfalse(filter_func, self))
        
        # Possibly panic if the list is empty.
        if not allow_empty and len(found) == 0:
            # Work out what we searched for to give better error.
            if label is not None:
                criteria_string = "label = '{}'".format(label)
            
            elif HOMO_difference is not None:
                criteria_string = "HOMO difference = '{}'".format(HOMO_difference)
            
            elif level is not None:
                criteria_string = "index = '{}'".format(level)
            
            raise Result_unavailable_error("Orbital", "could not find orbital with {}".format(criteria_string))
        
        return found
    
    def ordered(self):
        """
        Return a copy of this list of MOs that is ordered in terms of energy and removes duplicate MOs.
        """
        ordered_list = type(self)(set(self))
        #ordered_list.sort(key = lambda mo: mo.level)
        ordered_list.sort()
        return ordered_list
        
    @classmethod
    def from_parser(self, parser, cls = None):
        """
        Construct a Molecular_orbital_list object from an output file parser.
        
        :param parser: An output file parser.
        :param cls: Optional class of objects to populate this list with, should inherit from Molecular_orbital. Defaults to Molecular_orbital if only one set of orbitals are available, or Alpha_orbital if both alpha and beta are available (in which case you should call Molecular_orbital_list.from_cclib() again with cls = Beta_orbital to get beta as well).
        :returns: The new Molecular_orbital_list object. The list will be empty if no MO data is available.
        """
        try:
            # Set our default class if we've not been given one.
            if cls is None:
                # Check to see if we have only 'alpha' or beta as well.
                if len(parser.data.moenergies) == 1:
                    cls = Molecular_orbital
                else:
                    cls = Alpha_orbital
            # Get our list.
            return self(cls.list_from_parser(parser))
        except AttributeError:
            return self()
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        # Decide which class to use.
        if data['spin_type'] == "none":
            cls = Molecular_orbital
        elif data['spin_type'] == "alpha":
            cls = Alpha_orbital
        elif data['spin_type'] == "beta":
            cls = Beta_orbital
        elif data['spin_type'] is None and len(data['values']) == 0:
            # No orbitals, no type defined.
            return self()
        
        return self(cls.list_from_dump(data['values'], result_set, options))
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dict = {
            "dE(HOMO-LUMO)": {
                "value": self.HOMO_LUMO_energy if self.safe_get("HOMO_LUMO_energy") else None,
                "units": "eV"
            },
            "num_occupied": len(self.occupied),
            "num_virtual": len(self.virtual),
            "values": super()._dump_(digichem_options, all),
            "spin_type": self.safe_get("spin_type")
        }
#         # Add HOMO and LUMO
#         for label in ["HOMO", "LUMO"]:
#             try:
#                 orbital = self.find(label)
#                 dump_dict[label] = {
#                     "value": float(orbital.energy),
#                     "units": "eV"
#                 }
#                 
#             except Result_unavailable_error:
#                 pass
        
        return dump_dict
    
    def find_common_level(self, *other_lists, HOMO_difference):
        """
        Find either:
            The orbital with the lowest level that has no less than the given negative HOMO_difference.
        or:
            The orbital with the highest level that has no more than the given positive HOMO_difference.
        Across one or more orbital lists.
        
        The method is useful for determining which orbitals to traverse between two limits from the HOMO-LUMO gap.
        
        :raises Result_unavailable_error: If all of the given orbital_lists (including this one) are empty.
        :param *other_lists: Optional lists to search. If none are given, then only this orbital_list is search.
        :param HOMO_difference: The distance from the HOMO to search for. Negative values indicate HOMO-n, positive values indicate LUMO+(n-1). The LUMO should be at +1 by definition.
        :return: The level (as an integer) of the matching orbital.
        """
        # Our list of orbitals that match our criteria.
        found_orbitals = []
        
        # Our list of orbital_list objects to look through.
        orbital_lists = list(other_lists)
        orbital_lists.append(self)
        
        # Cant think of a better way to do this...
        if HOMO_difference <= 0:
            search_func = lambda orbital: orbital.HOMO_difference < HOMO_difference
        else:
            search_func = lambda orbital: orbital.HOMO_difference > HOMO_difference
        
        # Loop through each list and search.
        for orbital_list in orbital_lists:
            # Now search each list for orbitals that match our criteria.
            matching_orbitals = list(filterfalse(search_func, orbital_list))
            
            # We can just add all the orbitals we find because we'll only look at the lowest/highest anyway.
            found_orbitals.extend(matching_orbitals)
            
        # Get a list of orbital levels that match our criteria.
        orbital_levels = [orbital.level for orbital in found_orbitals]
        
        # Now either return the smallest or largest orbital level, depending on what we were asked for.
        try:
            if HOMO_difference <= 0:
                # The lowest orbital below HOMO.
                return min(orbital_levels)
            else:
                # The highest orbital above LUMO.
                return max(orbital_levels)
        except ValueError:
            # Min/Max couldn't find anything, this should only happen if all orbital_lists are completely empty.
            raise Result_unavailable_error("Common orbital level", "there are no orbitals")
    
    def assign_total_level(self, other_list):
        """
        Assign total levels to the orbitals in this list and another orbital list.
        
        'Total levels' are the index +1 of each orbital out of both alpha and beta orbitals.
        
        :param: other_list: If this list contains alpha orbitals, other_list should contain beta_orbitals (vice versa).
        """
        for index, orbital in enumerate(sorted(chain(self, other_list), key = lambda orbital: orbital.energy)):
            orbital.total_level = index +1
            
    @classmethod
    def merge_orbitals(self, molecular_orbital_lists, beta_orbital_lists):
        """
        """
        if len(molecular_orbital_lists) != len(beta_orbital_lists):
            # Panic.
            raise TypeError("molecular_orbital_lists and beta_orbital_lists must be of the same length")
        
        MOs = molecular_orbital_lists[0]
        betas = beta_orbital_lists[0]
        
        # Check all other lists are the same.
        for MO_list, beta_list in zip(molecular_orbital_lists[1:], beta_orbital_lists[1:]):
            # If this list has atoms and our current doesn't, use this as our base list.
            # We only do this if both lists are empty, so we don't have alpha and beta orbitals from two different results.
            if len(MOs) == 0 and len(betas) == 0:
                if len(MO_list) > 0:
                    MOs = MO_list
                if len(beta_list) > 0:
                    betas = beta_list
                
            else:
                MOs.check_equivalent(MOs, MO_list)
                betas.check_equivalent(betas, beta_list)
                
        return (MOs, betas)
    
    @classmethod
    def check_equivalent(self, MOs, other_MO_list):
        """
        Check that two MO lists are equivalent, issuing merge warnings if not.
        """
        for index, MO in enumerate(other_MO_list):
            try:
                other_MO = MOs[index]
                if not self.are_items_equal(MO, other_MO):
                    warnings.warn(self.MERGE_WARNING)
            except IndexError:
                warnings.warn(self.MERGE_WARNING)
        
            
    @classmethod
    def merge(self, *multiple_lists):
        """
        Merge multiple lists of MOs into a single list.
        """
        raise NotImplementedError("Molecular orbital lists do not implement the merge() method; use merge_orbitals() instead")
    
    @classmethod
    def are_items_equal(self, MO, other_MO):
        """
        A method which determines whether two items are the same for the purposes of merging.
        """
        return math.isclose(MO.energy, other_MO.energy) and MO.level == other_MO.level and MO.spin_type == other_MO.spin_type
    

class Molecular_orbital(Result_object, Floatable_mixin):
    """
    Class representing a molecular orbital.
    """
    
    # True MOs don't have a spin.
    spin_type = "none"
    
    def __init__(self, level, HOMO_difference, energy, symmetry = None, symmetry_level = None):
        """
        Constructor for MOs.
        
        :param level: The 'level' of this MO (essentially an index), where the lowest MO has a level of 1, increasing by 1 for each higher orbital.
        :param HOMO_difference: The distance of this MO from the HOMO. A negative value means this orbital is HOMO-n. A positive value means this orbital is HOMO+n (or LUMO+(n-1). A value of 0 means this orbital is the HOMO. A value of +1 means this orbital is the LUMO.
        :param energy: The energy of this MO (in eV).
        :param symmetry: The symmetry of this MO.
        :param symmetry_level: The ordered (by energy) index of this orbital in terms of orbitals that share the same symmetry.
        """
        super().__init__()
        self.level = level
        self.HOMO_difference = HOMO_difference
        self.symmetry = symmetry
        self.symmetry_level = symmetry_level
        self.energy = energy
                
        # This needs to be set outside of this constructor, because it relies on multiple lists of orbitals being completed.
        # Total level is the index +1 of this orbital out of both alpha and beta orbitals (for restricted this will == level).
        self.total_level = None
        
    @property
    def is_occupied(self):
        """
        Determine whether this orbital is occupied.
        """
        return self.HOMO_difference <= 0
            
    def __float__(self):
        return float(self.energy)
    
    @property
    def HOMO_level(self):
        """
        The level of the HOMO in the collection of orbitals of which this orbital is a member.
        """
        return self.level - self.HOMO_difference
    
    @property
    def LUMO_level(self):
        """
        The level of the LUMO in the collection of orbitals of which this orbital is a member.
        """
        return self.HOMO_level +1        
    
    @property
    def label(self):
        """
        A label describing this MO in terms of its proximity to the HOMO and LUMO.
        
        :return: A string label, of the form either HOMO-n or LUMO+n.
        """
        # The label we return depends on how close to the HOMO we are.
        if self.level == self.HOMO_level:
            # We are the HOMO.
            label = "HOMO"
        elif self.level < self.HOMO_level:
            # We are below the HOMO (and presumably occupied).
            label = "HOMO{}".format(self.level - self.HOMO_level)
        elif self.level == self.LUMO_level:
            # We are the LUMO.
            label = "LUMO"
        else:
            # We are above the LUMO (and presumably unoccupied).
            label = "LUMO{0:+}".format(self.level - self.LUMO_level)
            
        return label
    
    @property
    def irrep(self):
        """
        A unique description of this orbital combining symmetry and energy, as used by Turbomole.
        """
        return "{}{}".format(self.symmetry_level, self.symmetry.lower()) if self.symmetry_level is not None and self.symmetry is not None else None
    
    @property
    def sirrep(self):
        """
        A unique description of this orbital combining symmetry, energy and spin (alpha/beta), as used by Turbomole.
        """
        if self.spin_type == "none":
            return self.irrep
        else:
            if self.spin_type == "alpha":
                spin_tag = "a"
            elif self.spin_type == "beta":
                spin_tag = "b"
            else:
                spin_tag = "?"
            return "{}_{}".format(self.irrep, spin_tag)
    
    def __eq__(self, other):
        """
        Equality operator between MOs.
        """
        return self.label == other.label
    
    def __hash__(self):
        """
        Hash operator.
        """
        return hash(tuple(self.label))
    
    # The index used to access data from cclib (which always has two lists, one for alpha one for beta).
    ccdata_index = 0
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "index": self.level,
            "label": self.label,
            "homo_distance": int(self.HOMO_difference),
            "energy": {
                "value": float(self.energy),
                "units": "eV"
            },
            "symmetry": self.symmetry,
            "symmetry_index": self.symmetry_level
        }
        
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return [self(orbital_dict['index'], orbital_dict['homo_distance'], orbital_dict['energy']['value'], orbital_dict['symmetry'], orbital_dict['symmetry_index']) for orbital_dict in data]
    
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a list of Molecular_orbital objects from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of Molecular_orbital objects. The list will be empty if no MO is available.
        """
        try:
            # Get and zip required params.
            if hasattr(parser.data, "mosyms"):
                # We have symmetries.
                symmetry = parser.data.mosyms[self.ccdata_index]
            else:
                # No symmetry.
                symmetry = []
                
            # Don't catch this exception; if we don't have MO energies there's nothing we can do.
            energy = parser.data.moenergies[self.ccdata_index]
            
            # Check we don't have more symmetries than we do energies.
            if len(energy) < len(symmetry):
                diff = len(symmetry) - len(energy)
                warnings.warn("Parsed more MO symmetries than MO energies; excess symmetries will be discarded ('{}' symmetries, '{}' energies)".format(len(symmetry), len(energy)))
                # Dump the excess symmetry.
                symmetry = symmetry[:-diff]
            
            # Keep a track of all the symmetries we've seen.
            symmetries = {}

            orbitals = []
            for index, (symmetry, energy) in enumerate(zip_longest(symmetry, energy, fillvalue = None)):
                if symmetry is not None:
                    try:
                        # Add one to the level.
                        symmetries[symmetry] += 1
                        symm_level = symmetries[symmetry]
                    except KeyError:
                        # We haven't seen this mult before.
                        symmetries[symmetry] = 1
                        symm_level = symmetries[symmetry]
                else:
                    symm_level = None
                    
                orbitals.append(
                    self(
                        index +1,
                        index - parser.data.homos[self.ccdata_index],
                        energy,
                        symmetry,
                        symm_level
                    )
                )
                 
            return orbitals       
                
        except (AttributeError, IndexError):
            return []
    
class Unrestricted_orbital(Molecular_orbital):
    """
    Top-level class for unrestricted orbitals.
    """
    
    def __init__(self, level, HOMO_difference, energy, spin_type, symmetry = None, symm_level = None):
        """
        Constructor for MOs.
        
        :param level: The 'level' of the MO (essentially an index), where the lowest MO has a level of 1, increasing by 1 for each higher orbital.
        :param HOMO_difference: The distance of this MO from the HOMO. A negative value means this orbital is HOMO-n. A positive value means this orbital is HOMO+n (or LUMO+(n-1). A value of 0 means this orbital is the HOMO. A value of +1 means this orbital is the LUMO.
        :param symmetry: The symmetry of the MO.
        :param energy: The energy of the MO (in eV).
        :param spin_type: The spin of this spin-orbital (either alpha or beta).
        """
        # Call parent first.
        super().__init__(level, HOMO_difference, energy, symmetry, symm_level)
        self.spin_type = spin_type
    
    @property
    def label(self):
        # Get the base of the label first.
        label = super().label
        # Append our spin type.
        label = "{} ({})".format(label, self.spin_type)
        return label

class Alpha_orbital(Unrestricted_orbital):
    """
    An alpha spin orbital (these types of orbitals are only singly occupied, electrons are spin-up).
    """
    
    def __init__(self, level, HOMO_difference, energy, symmetry, symm_level):
        """
        Constructor for alpha MOs.
        
        :param level: The 'level' of the MO (essentially an index), where the lowest MO has a level of 1, increasing by 1 for each higher orbital.
        :param HOMO_difference: The distance of this MO from the HOMO. A negative value means this orbital is HOMO-n. A positive value means this orbital is HOMO+n (or LUMO+(n-1). A value of 0 means this orbital is the HOMO. A value of +1 means this orbital is the LUMO.
        :param energy: The energy of the MO (in eV).
        :param symmetry: The symmetry of the MO.
        """
        super().__init__(level, HOMO_difference, energy, "alpha", symmetry, symm_level)
        
class Beta_orbital(Unrestricted_orbital):
    """
    A beta spin orbital (these types of orbitals are only singly occupied, electrons are spin-down).
    """
    
    # Beta orbitals use the other list in cclib.
    ccdata_index = 1
    
    def __init__(self, level, HOMO_difference, energy, symmetry, symm_level):
        """
        Constructor for beta MOs.
        
        :param level: The 'level' of the MO (essentially an index), where the lowest MO has a level of 1, increasing by 1 for each higher orbital.
        :param HOMO_difference: The distance of this MO from the HOMO. A negative value means this orbital is HOMO-n. A positive value means this orbital is HOMO+n (or LUMO+(n-1). A value of 0 means this orbital is the HOMO. A value of +1 means this orbital is the LUMO.
        :param energy: The energy of the MO (in eV).
        :param symmetry: The symmetry of the MO.
        """
        super().__init__(level, HOMO_difference, energy, "beta", symmetry, symm_level)
    


from pathlib import Path

import digichem
from digichem.datas import get_resource


# A config file included in the digichem install directory.
master_config_path = get_resource("data/config/digichem.yaml")

# The location of the system-wide config file (which has precedence over the master).
system_config_location = Path("/etc/digichem{}/digichem.yaml".format(digichem.major_version))

# The location of the user specific config file (which has precedence over the system).
user_config_location = Path(Path.home(), ".config/digichem{}/digichem.yaml".format(digichem.major_version))


# from .base import Config
from .base import Digichem_options, Auto_type
from .locations import master_config_path, user_config_location, system_config_location
from .parse import Config_file_parser
from .util import get_config